"""
Tagulous test app views
"""
from __future__ import unicode_literals

from django.utils import six
from django.views.generic.edit import CreateView, UpdateView

# Django 1.10 deprecates urlresolvers
try:
    from django.urls import reverse_lazy
except ImportError:
    from django.core.urlresolvers import reverse_lazy

from tests.tagulous_tests_app import models


def null(request):
    "Null view for reversing"
    return None

class MixedCreate(CreateView):
    model = models.SimpleMixedTest
    fields = ['name', 'singletag', 'tags']
    success_url = reverse_lazy('tagulous_tests_app-null')

class MixedUpdate(UpdateView):
    model = models.SimpleMixedTest
    fields = ['name', 'singletag', 'tags']
    success_url = reverse_lazy('tagulous_tests_app-null')



"""
Tagulous test: Model SingleTagField

Modules tested:
    tagulous.models.managers.SingleTagManager
    tagulous.models.descriptors.BaseTagDescriptor
    tagulous.models.descriptors.SingleTagDescriptor
    tagulous.models.fields.BaseTagField
    tagulous.models.fields.SingleTagField
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from django.utils import six

from tests.lib import *


class ModelSingleTagFieldTest(TagTestManager, TestCase):
    """
    Test model SingleTagField
    """
    manage_models = [
        test_models.SingleTagFieldModel,
    ]

    def setUpExtra(self):
        self.tag_model = test_models.SingleTagFieldModel.title.tag_model
        self.tag_field = test_models.SingleTagFieldModel.title

    def test_descriptor(self):
        "Check SingleTagDescriptor is in place"
        self.assertIsInstance(
            self.tag_field, tag_models.SingleTagDescriptor
        )

    def test_tag_table(self):
        "Check the tag table exists"
        self.assertTrue(issubclass(
            self.tag_model, tag_models.TagModel
        ))

    def test_empty_value(self):
        "Check the descriptor returns None for no value"
        t1 = self.create(test_models.SingleTagFieldModel, name="Test")
        self.assertInstanceEqual(t1, name="Test", title=None)
        self.assertTagModel(self.tag_model, {})

    def test_tag_assign_in_constructor(self):
        "Check a tag string can be set in the constructor"
        t1 = test_models.SingleTagFieldModel(name="Test", title='Mr')
        t1.save()
        self.assertEqual(t1.name, 'Test')
        self.assertEqual(t1.title.name, 'Mr')
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
        })

    def test_assign_by_object_in_constructor(self):
        "Check a tag object can be passed in the constructor"
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1', title='Mr')
        t2 = test_models.SingleTagFieldModel(name='Test 2', title=t1.title)
        t2.save()
        self.assertEqual(t1.title, t2.title)
        self.assertEqual(t1.title.pk, t2.title.pk)
        self.assertTagModel(self.tag_model, {
            'Mr': 2,
        })

    def test_tag_assign_in_object_create(self):
        "Check a tag string can be passed in object.create"
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test', title='Mr')
        self.assertEqual(t1.name, 'Test')
        self.assertEqual(t1.title.name, 'Mr')
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
        })

    def test_assign_by_object_in_object_create(self):
        "Check a tag object can be passed in object.create"
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1', title='Mr')
        t2 = test_models.SingleTagFieldModel.objects.create(name='Test 2', title=t1.title)
        self.assertEqual(t1.title, t2.title)
        self.assertEqual(t1.title.pk, t2.title.pk)
        self.assertTagModel(self.tag_model, {
            'Mr': 2,
        })

    def test_tag_assign_in_object_get_or_create_true(self):
        """
        Check a tag string can be passed in object.get_or_create, when object
        does not exist
        """
        t1, state = test_models.SingleTagFieldModel.objects.get_or_create(
            name='Test', title='Mr',
        )
        self.assertEqual(state, True)
        self.assertEqual(t1.name, 'Test')
        self.assertEqual(t1.title.name, 'Mr')
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
        })

    def test_tag_assign_in_object_get_or_create_false(self):
        """
        Check a tag string can be passed in object.get_or_create, when object
        does exist
        """
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test', title='Mr')
        t2, state = test_models.SingleTagFieldModel.objects.get_or_create(
            name='Test', title='Mr',
        )
        self.assertEqual(state, False)
        self.assertEqual(t1.pk, t2.pk)
        self.assertEqual(t1.name, t2.name)
        self.assertEqual(t2.name, 'Test')
        self.assertEqual(t2.title.name, 'Mr')
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
        })

    def test_tag_assign_by_string(self):
        "Check a tag string can be assigned to descriptor and returned"
        t1 = self.create(test_models.SingleTagFieldModel, name="Test")
        t1.title = 'Mr'

        # Returned before save
        self.assertEqual(t1.title.__class__, self.tag_model)
        self.assertEqual(t1.title.name, 'Mr')
        self.assertEqual(six.text_type(t1.title), 'Mr')
        self.assertTagModel(self.tag_model, {})

        # Returned after save
        t1.save()
        self.assertEqual(t1.title.__class__, self.tag_model)
        self.assertEqual(t1.title.name, 'Mr')
        self.assertEqual(six.text_type(t1.title), 'Mr')
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
        })

    def test_assign_by_object(self):
        """
        Check a tag object can be assigned to a SingleTagfield, and that its
        tag count is incremented
        """
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1', title='Mr')
        t2 = test_models.SingleTagFieldModel(name='Test 2')
        t2.title = t1.title
        t2.save()
        self.assertEqual(t1.name, 'Test 1')
        self.assertEqual(t2.name, 'Test 2')
        self.assertEqual(six.text_type(t1.title), 'Mr')
        self.assertEqual(six.text_type(t2.title), 'Mr')
        self.assertEqual(t1.title, t2.title)
        self.assertEqual(t1.title.pk, t2.title.pk)
        self.assertTagModel(self.tag_model, {
            'Mr': 2,
        })

    def test_assign_string_empty(self):
        "Check an empty string can clear a SingleTagField"
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1', title='Mr')
        self.assertInstanceEqual(t1, name='Test 1', title='Mr')
        t1.title = ''
        t1.save()
        self.assertInstanceEqual(t1, name='Test 1', title=None)
        self.assertTagModel(self.tag_model, {})

    def test_assign_none(self):
        "Check assigning None can clear a SingleTagField"
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1', title='Mr')
        self.assertInstanceEqual(t1, name='Test 1', title='Mr')
        t1.title = None
        t1.save()
        self.assertInstanceEqual(t1, name='Test 1', title=None)
        self.assertTagModel(self.tag_model, {})

    def test_assign_string_quotes(self):
        "Check that a tag can contain quotes"
        # Check it saves ok
        t1 = test_models.SingleTagFieldModel.objects.create(
            name='Test 1', title='"One", Two"',
        )
        self.assertInstanceEqual(t1, name='Test 1', title='"One", Two"')

        # Check it loads ok
        t2 = test_models.SingleTagFieldModel.objects.get(pk=t1.pk)
        self.assertInstanceEqual(t2, name='Test 1', title='"One", Two"')

    def test_tag_assign_same(self):
        "Check that setting the same tag doesn't make a change"
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1', title='Mr')
        tag = t1.title
        self.assertEqual(tag, 'Mr')

        # Access the manager via the cache to watch changed state
        self.assertIsInstance(t1._title_tagulous, tag_models.SingleTagManager)
        self.assertFalse(t1._title_tagulous.changed, 'tag has changed')
        t1.title = 'Mr'
        self.assertFalse(t1._title_tagulous.changed, 'tag has changed')

    def test_change_decreases_count(self):
        "Check a tag string changes the count"
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1', title='Mr')
        t2 = test_models.SingleTagFieldModel.objects.create(name='Test 2', title=t1.title)
        self.assertTagModel(self.tag_model, {
            'Mr': 2,
        })
        t2.title = 'Mrs'
        t2.save()
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
            'Mrs': 1,
        })

    def test_delete_decreases_count(self):
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1', title='Mr')
        t2 = test_models.SingleTagFieldModel.objects.create(name='Test 2', title=t1.title)
        self.assertTagModel(self.tag_model, {
            'Mr': 2,
        })
        t2.delete()
        self.assertInstanceEqual(t1, name='Test 1', title='Mr')
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
        })

    def test_count_zero_deletes_tag(self):
        "Check a count of 0 deletes an unprotected tag"
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1', title='Mr')
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
        })
        t1.title = 'Mrs'
        t1.save()
        self.assertTagModel(self.tag_model, {
            'Mrs': 1,
        })

    def test_delete_decreases_correct(self):
        """
        Check that the actual tag in the database is decreased, not the one in
        the instance at time of deletion
        """
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1', title='Mr')
        t2 = test_models.SingleTagFieldModel.objects.create(name='Test 2', title='Mrs')
        self.assertEqual(six.text_type(t1.title.name), 'Mr')
        self.assertEqual(six.text_type(t2.title.name), 'Mrs')
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
            'Mrs': 1,
        })

        # Now change the title and delete without saving
        t1.title = 'Mrs'
        self.assertEqual(six.text_type(t1.title.name), 'Mrs')
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
            'Mrs': 1,
        })
        t1.delete()

        # Check the original tag 'Mr' was decremented (and deleted)
        self.assertTagModel(self.tag_model, {
            'Mrs': 1,
        })

        # But check that tagulous still thinks the tag is 'Mrs'
        self.assertEqual(six.text_type(t1.title.name), 'Mrs')

    def test_save_deleted_instance(self):
        """
        Check that a deleted tag in memory can be re-saved when the instance it
        is set on is deleted
        """
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1', title='Mr')
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
        })
        t1.delete()
        self.assertTagModel(self.tag_model, {})
        t1.save()
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
        })

    def test_save_deleted_tag(self):
        """
        Check that a delete tag in memory can be read and re-saved when it is
        deleted without the instance knowing about it
        """
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1', title='Mr')
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
        })
        self.tag_model.objects.all().delete()
        self.assertTagModel(self.tag_model, {})

        # Check it's still usable
        self.assertIsInstance(t1.title, tag_models.BaseTagModel)
        self.assertEqual(t1.title, 'Mr')

        # Check it can be re-saved
        t1.save()
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
        })

    def test_multiple_unsaved(self):
        "Check that there's no leak between unsaved objects"
        t1 = test_models.SingleTagFieldModel(name='Test 1', title='Mr')
        t2 = test_models.SingleTagFieldModel(name='Test 1', title='Mrs')
        self.assertTagModel(self.tag_model, {})
        self.assertEqual(six.text_type(t1.title), 'Mr')
        self.assertEqual(six.text_type(t2.title), 'Mrs')
        t1.save()
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
        })
        t2.save()
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
            'Mrs': 1,
        })
        self.assertEqual(six.text_type(t1.title), 'Mr')
        self.assertEqual(six.text_type(t2.title), 'Mrs')

    def test_load_instance(self):
        "Check that SingleTagField is loaded correctly"
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1', title='Mr')
        t2 = test_models.SingleTagFieldModel.objects.get(pk=t1.pk)
        self.assertIsInstance(t2.title, tag_models.BaseTagModel)
        self.assertEqual(t1.title, t2.title)
        self.assertTagModel(self.tag_model, {
            'Mr': 1,
        })

    def test_descriptor_equal(self):
        "Check that descriptors evaluate to equal"
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1', title='Mr')
        t2 = test_models.SingleTagFieldModel.objects.create(name='Test 2', title='Mr')
        self.assertIsInstance(t1.title, self.tag_model)
        self.assertEqual(t1.title.pk, t2.title.pk)
        self.assertEqual(t1.title, t2.title)

    def test_descriptor_not_equal(self):
        "Check that descriptors evaluate to equal"
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1', title='Mr')
        t2 = test_models.SingleTagFieldModel.objects.create(name='Test 2', title='Mrs')
        self.assertIsInstance(t1.title, self.tag_model)
        self.assertNotEqual(t1.title, t2.title)

    def test_cascade_delete(self):
        "Check that deleting a tag deletes its related tagged items (by default)"
        model = test_models.SingleTagFieldModel
        t1 = model.objects.create(name='Test 1', title='Mr')
        t2 = model.objects.create(name='Test 2', title='Mr')
        t3 = model.objects.create(name='Test 3', title='Mrs')
        self.assertTagModel(self.tag_model, {
            'Mr': 2,
            'Mrs': 1,
        })
        self.assertSequenceEqual(model.objects.all(), [t1, t2, t3])

        # Delete
        self.tag_model.objects.get(name='Mr').delete()
        self.assertTagModel(self.tag_model, {
            'Mrs': 1,
        })
        self.assertSequenceEqual(model.objects.all(), [t3])


###############################################################################
#######  Test invalid fields
###############################################################################

class ModelSingleTagFieldInvalidTest(TagTestManager, TransactionTestCase):
    """
    Test invalid model SingleTagField

    Use a TransactionTestCase so the apps will be reset
    """
    manage_models = [
        test_models.SingleTagFieldModel,
    ]

    def test_invalid_to_model(self):
        "Check that the to model has to be a TagModel subclass"
        with self.assertRaises(ValueError) as cm:
            class FailModel_invalid_to(models.Model):
                to_model = tag_models.SingleTagField(test_models.SingleTagFieldModel)
        self.assertEqual(
            six.text_type(cm.exception),
            "Tag model must be a subclass of TagModel"
        )

    def test_forbidden_to_field(self):
        "Check that to_field argument raises exception"
        with self.assertRaises(ValueError) as cm:
            class FailModel_forbidden_to(models.Model):
                to_field = tag_models.SingleTagField(to_field='fail')
        self.assertEqual(
            six.text_type(cm.exception),
            "Invalid argument 'to_field' for SingleTagField"
        )

    def test_forbidden_rel_class(self):
        "Check that rel_class argument raises exception"
        with self.assertRaises(ValueError) as cm:
            class FailModel_forbidden_rel(models.Model):
                rel_class = tag_models.SingleTagField(rel_class='fail')
        self.assertEqual(
            six.text_type(cm.exception),
            "Invalid argument 'rel_class' for SingleTagField"
        )

    def test_forbidden_max_count(self):
        "Check that max_count argument raises exception"
        with self.assertRaises(ValueError) as cm:
            class FailModel_forbidden_max_count(models.Model):
                max_count = tag_models.SingleTagField(max_count='fail')
        self.assertEqual(
            six.text_type(cm.exception),
            "Invalid argument 'max_count' for SingleTagField"
        )

    def test_value_from_object_none(self):
        "Check that value_from_object returns empty string instead of None"
        # Called by forms, but part of model, so test it here
        t1 = test_models.SingleTagFieldModel.objects.create(name='Test 1')
        field = test_models.SingleTagFieldModel._meta.get_field('title')
        self.assertEqual(field.value_from_object(t1), '')


###############################################################################
#######  Test model field blank=True
###############################################################################

class ModelSingleTagFieldOptionalTest(TagTestManager, TestCase):
    """
    Test optional model SingleTagField
    """
    manage_models = [
        test_models.SingleTagFieldOptionalModel,
    ]
    def test_optional_save_missing(self):
        "Check an optional SingleTagField isn't required for save"
        # If it fails, it will return an exception
        t1 = test_models.SingleTagFieldOptionalModel(name='Test 1')
        t1.save()
        self.assertNotEqual(t1.pk, None)

    def test_optional_create_missing(self):
        "Check an optional SingleTagField isn't required for object.create"
        # If it fails, it will return an exception
        t1 = test_models.SingleTagFieldOptionalModel.objects.create(name='Test 1')
        self.assertNotEqual(t1.pk, None)


###############################################################################
####### Test model field blank=False
###############################################################################

class ModelSingleTagFieldRequiredTest(TagTestManager, TestCase):
    """
    Test required model SingleTagField
    """
    manage_models = [
        test_models.SingleTagFieldRequiredModel,
    ]
    def test_required_save_raises(self):
        "Check a required SingleTagField raises an exception when saved"
        with self.assertRaises(exceptions.ValidationError) as cm:
            t1 = test_models.SingleTagFieldRequiredModel(name='Test')
            t1.save()
        self.assertEqual(cm.exception.messages[0], 'This field cannot be null.')

    def test_required_create_raises(self):
        "Check a required SingleTagField raises an exception in object.create"
        with self.assertRaises(exceptions.ValidationError) as cm:
            t1 = test_models.SingleTagFieldRequiredModel.objects.create(name='Test')
        self.assertEqual(cm.exception.messages[0], 'This field cannot be null.')


###############################################################################
####### Test multiple SingleTagFields on a model
###############################################################################

class ModelSingleTagFieldMultipleTest(TagTestManager, TestCase):
    """
    Test multiple tag fields on a model
    """
    manage_models = [
        test_models.SingleTagFieldMultipleModel,
    ]

    def setUpExtra(self):
        self.test_model = test_models.SingleTagFieldMultipleModel
        self.tag_field_1 = test_models.SingleTagFieldMultipleModel.tag1
        self.tag_field_2 = test_models.SingleTagFieldMultipleModel.tag2
        self.tag_field_3 = test_models.SingleTagFieldMultipleModel.tag3

    def test_tag_models(self):
        self.assertTrue(issubclass(self.tag_field_1.tag_model, tag_models.TagModel))
        self.assertTrue(issubclass(self.tag_field_2.tag_model, tag_models.TagModel))
        self.assertTrue(issubclass(self.tag_field_3.tag_model, tag_models.TagModel))

    def test_separate_models(self):
        self.assertNotEqual(self.tag_field_1.tag_model, self.tag_field_2.tag_model)
        self.assertNotEqual(self.tag_field_1.tag_model, self.tag_field_3.tag_model)
        self.assertNotEqual(self.tag_field_2.tag_model, self.tag_field_3.tag_model)

    def test_model_names(self):
        self.assertEqual(
            self.tag_field_1.tag_model.__name__,
            'Tagulous_SingleTagFieldMultipleModel_tag1'
        )
        self.assertEqual(
            self.tag_field_2.tag_model.__name__,
            'Tagulous_SingleTagFieldMultipleModel_tag2'
        )
        self.assertEqual(
            self.tag_field_3.tag_model.__name__,
            'Tagulous_SingleTagFieldMultipleModel_tag3'
        )

    def test_set_and_get(self):
        "Test multiple fields can be set and retrieved independently"
        t1 = self.create(
            self.test_model, name="Test 1",
            tag1='Mr', tag2='blue', tag3='adam',
        )
        t2 = self.create(
            self.test_model, name="Test 2",
            tag1='Mrs', tag2='green', tag3='brian',
        )
        t3 = self.create(
            self.test_model, name="Test 3",
            tag1='Ms', tag2='red', tag3='chris',
        )

        self.assertTagModel(self.tag_field_1, {
            'Mr':   1,
            'Mrs':  1,
            'Ms':   1,
        })
        self.assertTagModel(self.tag_field_2, {
            'blue': 1,
            'green': 1,
            'red':  1,
        })
        self.assertTagModel(self.tag_field_3, {
            'adam': 1,
            'brian': 1,
            'chris': 1,
        })

        self.assertInstanceEqual(t1, name="Test 1", tag1='Mr', tag2='blue', tag3='adam')
        self.assertInstanceEqual(t2, name="Test 2", tag1='Mrs', tag2='green', tag3='brian')
        self.assertInstanceEqual(t3, name="Test 3", tag1='Ms', tag2='red', tag3='chris')


###############################################################################
####### Test SingleTagField with string references to tag model
###############################################################################

class ModelSingleTagFieldStringTest(TagTestManager, TransactionTestCase):
    """
    Test SingleTagField which refers to its tag model with a string
    """
    manage_models = [
        test_models.MixedStringTo,
    ]

    def setUpExtra(self):
        self.test_model = test_models.MixedStringTo
        self.tag_field = self.test_model.singletag
        self.tag_model = test_models.MixedStringTagModel

    def test_to_model(self):
        "Check related model is correct"
        self.assertTrue(issubclass(self.tag_field.tag_model, tag_models.TagModel))
        if django.VERSION < (1, 9):
            self.assertEqual(self.tag_field.field.remote_field.to, self.tag_model)
        else:
            self.assertEqual(self.tag_field.field.remote_field.model, self.tag_model)
        self.assertEqual(self.tag_field.tag_model, self.tag_model)

    def test_tag_options(self):
        "Check tag options are available correctly"
        self.assertEqual(
            self.tag_field.tag_options, self.tag_model.tag_options
        )

    def test_use(self):
        "Test basic use of tag field"
        self.assertTagModel(self.tag_model, {})
        t1 = self.test_model.objects.create(name='Test 1', singletag='Mr')
        self.assertTagModel(self.tag_model, {
            'Mr':  1,
        })
        self.assertInstanceEqual(t1, name='Test 1', singletag='Mr')


class ModelSingleTagFieldSelfTest(TagTestManager, TransactionTestCase):
    """
    Test SingleTagField which refers to itself
    """
    manage_models = [
        test_models.MixedSelfTo,
    ]

    def setUpExtra(self):
        self.test_model = test_models.MixedSelfTo
        self.tag_field = self.test_model.alternate

    def test_to_model(self):
        "Check related model is correct"
        self.assertTrue(issubclass(self.tag_field.tag_model, tag_models.TagModel))
        if django.VERSION < (1, 9):
            self.assertEqual(self.tag_field.field.remote_field.to, self.test_model)
        else:
            self.assertEqual(self.tag_field.field.remote_field.model, self.test_model)
        self.assertEqual(self.tag_field.tag_model, self.test_model)

    def test_tag_options(self):
        "Check tag options are available correctly"
        self.assertEqual(
            self.tag_field.tag_options, self.test_model.tag_options
        )

    def test_use(self):
        "Test basic use of tag field"
        self.assertTagModel(self.test_model, {})
        t1 = self.test_model.objects.create(name='Test 1', alternate='mr')
        self.assertTagModel(self.test_model, {
            'mr':  1,
            'Test 1': 0,
        })
        self.assertInstanceEqual(t1, name='Test 1', alternate='mr')


###############################################################################
####### Test SingleTagField options
###############################################################################

class ModelSingleTagFieldOptionsTest(TagTestManager, TestCase):
    """
    Test single tag field options
    """
    manage_models = [
        test_models.SingleTagFieldOptionsModel,
    ]

    def setUpExtra(self):
        self.test_model = test_models.SingleTagFieldOptionsModel

    def test_initial_string(self):
        # Initial will have been loaded by TagTestManager
        self.assertTagModel(self.test_model.initial_string, {
            'Mr':   0,
            'Mrs':  0,
            'Ms':   0,
        })

    def test_initial_list(self):
        # Initial will have been loaded by TagTestManager
        self.assertTagModel(self.test_model.initial_list, {
            'Mr':   0,
            'Mrs':  0,
            'Ms':   0,
        })

    def test_protect_initial_true(self):
        self.assertTagModel(self.test_model.protect_initial_true, {
            'Mr':   0,
        })

        t1 = self.create(self.test_model, name="Test 1", protect_initial_true='Mr')
        self.assertTagModel(self.test_model.protect_initial_true, {
            'Mr':   1,
        })

        t1.protect_initial_true = ''
        t1.save()
        self.assertTagModel(self.test_model.protect_initial_true, {
            'Mr':   0,
        })

    def test_protect_initial_false(self):
        self.assertTagModel(self.test_model.protect_initial_true, {
            'Mr':   0,
        })

        t1 = self.create(self.test_model, name="Test 1", protect_initial_false='Mr')
        self.assertTagModel(self.test_model.protect_initial_false, {
            'Mr':   1,
        })

        t1.protect_initial_false = ''
        t1.save()
        self.assertTagModel(self.test_model.protect_initial_false, {})

    def test_protect_all_true(self):
        t1 = self.create(self.test_model, name="Test 1", protect_all_true='Mr')
        self.assertTagModel(self.test_model.protect_all_true, {
            'Mr':   1,
        })

        t1.protect_all_true = ''
        t1.save()
        self.assertTagModel(self.test_model.protect_all_true, {
            'Mr':   0,
        })

    def test_protect_all_false(self):
        t1 = self.create(self.test_model, name="Test 1", protect_all_false='Mr')
        self.assertTagModel(self.test_model.protect_all_false, {
            'Mr':   1,
        })

        t1.protect_all_false = ''
        t1.save()
        self.assertTagModel(self.test_model.protect_all_false, {})

    @skip_if_mysql
    def test_case_sensitive_true(self):
        self.assertTagModel(self.test_model.case_sensitive_true, {
            'Mr':   0,
        })
        t1 = self.create(self.test_model, name="Test 1", case_sensitive_true='mr')
        self.assertTagModel(self.test_model.case_sensitive_true, {
            'Mr':   0,
            'mr':   1,
        })

    @skip_if_mysql
    def test_case_sensitive_false(self):
        self.assertTagModel(self.test_model.case_sensitive_false, {
            'Mr':   0,
        })
        t1 = self.create(self.test_model, name="Test 1", case_sensitive_false='mr')
        self.assertTagModel(self.test_model.case_sensitive_false, {
            'Mr':   1,
        })

    def test_force_lowercase_true(self):
        t1 = self.create(self.test_model, name="Test 1", force_lowercase_true='Mr')
        self.assertTagModel(self.test_model.force_lowercase_true, {
            'mr':   1,
        })

    def test_force_lowercase_false(self):
        t1 = self.create(self.test_model, name="Test 1", force_lowercase_false='Mr')
        self.assertTagModel(self.test_model.force_lowercase_false, {
            'Mr':   1,
        })



"""
Tagulous test: View-related functionality

Modules tested:
    tagulous.views
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from tests.lib import *

from django.test import Client
from django.utils import six

# Django 1.4 is last to support Python 2.5, but json isn't available until 2.6
try:
    import json
except ImportError:
   from django.utils import simplejson as json

# Django 1.10 deprecates urlresolvers
try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

client = Client()


###############################################################################
####### Test tag forms with CBVs
###############################################################################

class TagFormCBVTest(TagTestManager, TestCase):
    "Test CBVs which use tagged forms"
    manage_models = [
        test_models.SimpleMixedTest,
    ]

    def setUpExtra(self):
        # Load initial tags for all models which have them
        self.model = test_models.SimpleMixedTest
        self.singletag_model = self.model.singletag.tag_model
        self.tags_model = self.model.tags.tag_model

    def test_cbv_create(self):
        "Test CBV create"
        self.assertTagModel(self.singletag_model, {})
        self.assertTagModel(self.tags_model, {})

        # Create
        response = client.post(reverse('tagulous_tests_app-MixedCreate'), {
            'name': 'Test 1',
            'singletag': 'Mr',
            'tags': 'blue, red',
        })
        self.assertEqual(response.status_code, 302)

        # Get
        t1 = self.model.objects.get(name='Test 1')
        self.assertIsInstance(t1.singletag, tag_models.BaseTagModel)
        self.assertEqual(t1.singletag.name, 'Mr')
        self.assertEqual(t1.tags.count(), 2)
        self.assertEqual(t1.tags.all()[0], 'blue')
        self.assertEqual(t1.tags.all()[1], 'red')
        self.assertTagModel(self.singletag_model, {
            'Mr':   1,
        })
        self.assertTagModel(self.tags_model, {
            'blue': 1,
            'red':  1,
        })

    def test_cbv_update(self):
        "Test CBV update"
        # Set up object
        t1 = self.model.objects.create(
            name='Test 1',
            singletag='Mr',
            tags='blue, red',
        )
        self.assertTagModel(self.singletag_model, {
            'Mr':   1,
        })
        self.assertTagModel(self.tags_model, {
            'blue': 1,
            'red':  1,
        })

        # Update it
        response = client.post(
            reverse('tagulous_tests_app-MixedUpdate', kwargs={'pk': t1.pk}),
            {
                'name': 'Test 2',
                'singletag': 'Mrs',
                'tags': 'blue, green',
            }
        )
        self.assertEqual(response.status_code, 302)

        # Check
        t2 = self.model.objects.get(pk=t1.pk)
        self.assertIsInstance(t1.singletag, tag_models.BaseTagModel)
        self.assertEqual(t2.name, 'Test 2')
        self.assertEqual(t2.singletag.name, 'Mrs')
        self.assertEqual(t2.tags.count(), 2)
        self.assertEqual(t2.tags.all()[0], 'blue')
        self.assertEqual(t2.tags.all()[1], 'green')
        self.assertTagModel(self.singletag_model, {
            'Mrs':  1,
        })
        self.assertTagModel(self.tags_model, {
            'blue': 1,
            'green': 1,
        })


###############################################################################
####### Test autocomplete view
###############################################################################

def get_response_content(response):
    return response.content.decode('utf-8')


class AutocompleteViewTest(TagTestManager, TestCase):
    "Test autocomplete view"
    manage_models = [
        test_models.TagFieldOptionsModel,
    ]

    def setUpExtra(self):
        # Load initial tags for all models which have them
        self.test_model = test_models.TagFieldOptionsModel

    def test_unlimited(self):
        "Test unlimited autocomplete view"
        # Add some tags
        tag_model = self.test_model.autocomplete_view.tag_model
        for i in range(100):
            tag_model.objects.create(name='tag%02d' % i)
        self.assertEqual(tag_model.objects.count(), 100)

        # Get them from view
        response = client.get(
            reverse('tagulous_tests_app-unlimited'),
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data['results']), 100)
        for i in range(100):
            self.assertEqual(data['results'][i], 'tag%02d' % i)
        self.assertEqual(data['more'], False)

    def test_unlimited_query(self):
        "Test unlimited autocomplete view with query"
        # Add some tags
        tag_model = self.test_model.autocomplete_view.tag_model
        for i in range(100):
            tag_model.objects.create(name='tag%02d' % i)
        self.assertEqual(tag_model.objects.count(), 100)

        # Get them from view
        response = client.get(
            reverse('tagulous_tests_app-unlimited'),
            {'q': 'tag0'},
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data['results']), 10)
        for i in range(10):
            self.assertEqual(data['results'][i], 'tag%02d' % i)
        self.assertEqual(data['more'], False)

    def test_limited(self):
        "Test limited autocomplete view"
        # Add some tags
        tag_model = self.test_model.autocomplete_limit.tag_model
        for i in range(100):
            tag_model.objects.create(name='tag%02d' % i)
        self.assertEqual(tag_model.objects.count(), 100)

        page_length = tag_model.tag_options.autocomplete_limit
        self.assertEqual(page_length, 3)

        # Get page 1
        response = client.get(
            reverse('tagulous_tests_app-limited'),
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data['results']), page_length)
        self.assertEqual(data['results'][0], 'tag00')
        self.assertEqual(data['results'][1], 'tag01')
        self.assertEqual(data['results'][2], 'tag02')
        self.assertEqual(data['more'], True)

        # Get page 4: starting 10th tag, tag09 to tag11
        page = 4
        response = client.get(
            reverse('tagulous_tests_app-limited'),
            {'p': page},
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data['results']), page_length)
        self.assertEqual(data['results'][0], 'tag09')
        self.assertEqual(data['results'][1], 'tag10')
        self.assertEqual(data['results'][2], 'tag11')
        self.assertEqual(data['more'], True)

        # Get last page, 34: starting 100th tag, tag99
        page = 34
        response = client.get(
            reverse('tagulous_tests_app-limited'),
            {'p': page},
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0], 'tag99')
        self.assertEqual(data['more'], False)

    def test_limited_query(self):
        "Test limited autocomplete view with query"
        # Add some tags
        tag_model = self.test_model.autocomplete_limit.tag_model
        for i in range(100):
            tag_model.objects.create(name="tag%02d" % i)
        self.assertEqual(tag_model.objects.count(), 100)

        page_length = tag_model.tag_options.autocomplete_limit
        self.assertEqual(page_length, 3)

        # Get page 1
        response = client.get(
            reverse('tagulous_tests_app-limited'),
            {'q': 'tag1'},
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data['results']), page_length)
        self.assertEqual(data['results'][0], 'tag10')
        self.assertEqual(data['results'][1], 'tag11')
        self.assertEqual(data['results'][2], 'tag12')
        self.assertEqual(data['more'], True)

        # Get last page, 4: starting 10th tag, tag19
        page = 4
        response = client.get(
            reverse('tagulous_tests_app-limited'),
            {'q': 'tag1', 'p': page},
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0], 'tag19')
        self.assertEqual(data['more'], False)

    def test_login(self):
        "Test autocomplete_login view"
        # Add some tags
        tag_model = self.test_model.autocomplete_view.tag_model
        for i in range(100):
            tag_model.objects.create(name='tag%02d' % i)
        self.assertEqual(tag_model.objects.count(), 100)

        # Get them from view
        user = User.objects.create_user('test', 'test@example.com', 'password')
        client.login(username='test', password='password')
        response = client.get(
            reverse('tagulous_tests_app-login'),
        )
        client.logout()
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data['results']), 100)
        for i in range(100):
            self.assertEqual(data['results'][i], 'tag%02d' % i)
        self.assertEqual(data['more'], False)

    def test_queryset(self):
        "Test autocomplete view on a tag model queryset"
        # Add some tags
        tag_model = self.test_model.autocomplete_view.tag_model
        for i in range(100):
            tag_model.objects.create(name='tag%02d' % i)
        self.assertEqual(tag_model.objects.count(), 100)

        # Get them from view
        response = client.get(
            reverse('tagulous_tests_app-queryset'),
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data['results']), 10)
        for i in range(10):
            self.assertEqual(data['results'][i], 'tag2%d' % i)
        self.assertEqual(data['more'], False)

    def test_force_lowercase_true(self):
        "Test autocomplete view on a tag model with force_lowercase=True"
        # Add some tags
        tag_model = self.test_model.force_lowercase_true.tag_model
        for i in range(100):
            tag_model.objects.create(name='tag%02d' % i)
        self.assertEqual(tag_model.objects.count(), 100)

        # Get them from view
        response = client.get(
            reverse('tagulous_tests_app-force_lowercase_true'),
            {'q': 'Tag1'},
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data['results']), 10)
        for i in range(10):
            self.assertEqual(data['results'][i], 'tag1%d' % i)
        self.assertEqual(data['more'], False)

    @skip_if_mysql
    def test_case_sensitive_false(self):
        "Test autocomplete view on a tag model with case_sensitive=False"
        # Add some tags
        tag_model = self.test_model.case_sensitive_false.tag_model
        tag_model.objects.all().delete() # clear out initial
        for i in range(100):
            tag_model.objects.create(name='tag%02d' % i)
        self.assertEqual(tag_model.objects.count(), 100)

        # Get them from view
        response = client.get(
            reverse('tagulous_tests_app-case_sensitive_false'),
            {'q': 'Tag1'},
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data['results']), 10)
        for i in range(10):
            self.assertEqual(data['results'][i], 'tag1%d' % i)
        self.assertEqual(data['more'], False)

    def test_case_sensitive_true(self):
        "Test autocomplete view on a tag model with case_sensitive=True"
        # Add some tags
        tag_model = self.test_model.case_sensitive_true.tag_model
        tag_model.objects.all().delete() # clear out initial
        for i in range(100):
            tag_model.objects.create(name='tag%02d' % i)
        self.assertEqual(tag_model.objects.count(), 100)

        # Get them from view
        response = client.get(
            reverse('tagulous_tests_app-case_sensitive_true'),
            {'q': 'Tag1'},
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))

        from django.db import connection
        if connection.vendor == 'sqlite':
            # Sqlite doesn't support case insensitive searches - expect to
            # get back 10 matches
            self.assertEqual(len(data['results']), 10)
            for i in range(10):
                self.assertEqual(data['results'][i], 'tag1%d' % i)
        else:
            self.assertEqual(len(data['results']), 0)
            self.assertEqual(data['more'], False)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import json
import re
import os
import sys
import unittest

import django
from django.db import connection, models
from django.contrib.auth.models import User
from django.core import exceptions
from django.test import TestCase, TransactionTestCase, testcases
from django.utils import six

from tagulous import constants as tag_constants
from tagulous import models as tag_models
from tagulous import forms as tag_forms
from tagulous import admin as tag_admin
from tagulous import utils as tag_utils
from tagulous import settings as tag_settings

from tests.tagulous_tests_app import models as test_models
from tests.tagulous_tests_app import admin as test_admin
from tests.tagulous_tests_app import urls as test_urls
from tests.tagulous_tests_app import forms as test_forms
from tests.tagulous_tests_app2 import models as test_models2

if six.PY3:
    from io import StringIO
else:
    from cStringIO import StringIO

# Detect test environment
# This is used when creating files (migrations and fixtures) to ensure that
# tests can be run concurrently without interfering with each other.
# Also calculated in setup.py
testenv = re.sub(
    r'[^a-zA-Z0-9]', '_',
    os.environ.get('TOXENV', '_'.join(str(v) for v in django.VERSION)),
)


class TagTestManager(object):
    """
    Test mixin to help test tag models
    """
    # We have some very long string comparisons (eg form field renders,
    # migrations), so set the maxDiff to 10k
    maxDiff = 1024 * 20

    # This class can manage models
    manage_models = None

    # Show normal fail messages in addition to any custom
    longMessage = True

    def setUp(self):
        """
        Ensure initial data is in the tag models
        """
        if self.manage_models is not None:
            for model in self.manage_models:
                tag_models.initial.model_initialise_tags(model)
                tag_models.initial.model_initialise_tags(model)

        if hasattr(self, 'setUpExtra'):
            self.setUpExtra()

    def tearDown(self):
        """
        Common tear down operations
        """
        # Only here for consistency
        if hasattr(self, 'tearDownExtra'):
            self.tearDownExtra()

    def create(self, model, **kwargs):
        # This could be replaced with the Tagulous create() method, but do it
        # manually to avoid complicating tests
        normal = {}
        tagfield = {}
        mmfield = {}
        for field_name, val in kwargs.items():
            if isinstance(
                model._meta.get_field(field_name), tag_models.TagField
            ):
                tagfield[field_name] = val
            elif isinstance(
                model._meta.get_field(field_name), models.ManyToManyField
            ):
                mmfield[field_name] = val
            else:
                normal[field_name] = val

        # Create as normal
        item = model.objects.create(**normal)

        # Add tagfields (may not be using enhanced queryset)
        for field_name, val in tagfield.items():
            setattr(item, field_name, val)
            getattr(item, field_name).save()

        # Add mm fields
        for field_name, val in mmfield.items():
            field = getattr(item, field_name)
            for obj in val:
                field.add(obj)

        return item

    def assertRegex(self, text, expected_regex, msg="Regex didn't match"):
        # Implement python 3.2's assertRegex for all
        if isinstance(text, six.string_types):
            expected_regex = re.compile(expected_regex)
        self.assertTrue(expected_regex.search(text), msg=msg)

    def assertInstanceEqual(self, instance, **kwargs):
        # First, reload instance
        instance = instance.__class__.objects.get(pk=instance.pk)

        # Check values
        for field_name, val in kwargs.items():
            try:
                if isinstance(
                    instance.__class__._meta.get_field(field_name),
                    (tag_models.SingleTagField, tag_models.TagField)
                ) and isinstance(val, six.string_types):
                    self.assertEqual(
                        six.text_type(getattr(instance, field_name)), val,
                    )
                elif isinstance(
                    instance.__class__._meta.get_field(field_name),
                    models.ManyToManyField
                ):
                    mm_objs = list(getattr(instance, field_name).all())
                    self.assertEqual(len(val), len(mm_objs))
                    for obj in val:
                        self.assertTrue(obj in mm_objs)
                else:
                    self.assertEqual(getattr(instance, field_name), val)
            except AssertionError as e:
                self.fail(
                    'Instances not equal for field %s: %s' % (field_name, e)
                )

    def assertTagModel(self, model, tag_counts):
        """
        Assert the tag model matches the specified tag counts
        """
        if isinstance(model, (tag_models.SingleTagDescriptor, tag_models.TagDescriptor)):
            model = model.tag_model

        if len(tag_counts) != model.objects.count():
            self.fail("Incorrect number of tags in '%s'; expected %d, got %d" % (model, len(tag_counts), model.objects.count()))

        for tag_name, count in tag_counts.items():
            try:
                tag = model.objects.get(name=tag_name)
            except model.DoesNotExist:
                self.fail("Tag model missing expected tag '%s'" % tag_name)
            if tag.count != count:
                self.fail("Tag count for '%s' incorrect; expected %d, got %d" % (tag_name, count, tag.count))

    def _extract_json(self, dom, path=''):
        "Recursively break out json from a django.utils.html_parser dom object"
        jsons = {}
        if isinstance(dom, six.string_types):
            return dom, {}
        el_name = path + '.' + dom.name

        # Extract json from this element
        for i, attr in enumerate(dom.attributes):
            attr_name, attr_val = attr
            if not attr_name.startswith('data-'):
                continue

            if (
                (attr_val[0] == '{' and attr_val[-1] == '}')
                or (attr_val[0] == '[' and attr_val[-1] == ']')
            ):
                if el_name not in jsons:
                    jsons[el_name] = {}
                jsons[el_name][attr_name] = attr_val
                dom.attributes[i] = (attr_name, '/*json*/')

        # Look at children
        for child in dom.children:
            new_child, new_jsons = self._extract_json(child, el_name)
            jsons.update(new_jsons)
        return dom, jsons

    def assertHTMLEqual(self, html1, html2, msg=None):
        """
        Clone of django's method, but with support for JSON in data- tag
        attributes and backwards-compatible support for HTML 5 "required"
        """
        # Add backwards compatibility for HTML 5 "required"
        REQUIRED = '{{required}}'
        required = ''
        if django.VERSION >= (1, 10):
            required = ' required '
        if REQUIRED in html1:
            html1 = html1.replace(REQUIRED, required)
        if REQUIRED in html2:
            html2 = html2.replace(REQUIRED, required)

        # Parse the HTML into a dom
        err1 = 'First argument is not a valid HTML element:'
        err2 = 'Second argument is not a valid HTML element:'
        dom1 = testcases.assert_and_parse_html(self, html1, msg, err1)
        dom2 = testcases.assert_and_parse_html(self, html2, msg, err2)

        # Check it actually managed to parse something
        self.assertNotEqual(dom1.name, None, msg='%s\n%s' % (err1, html1))
        self.assertNotEqual(dom2.name, None, msg='%s\n%s' % (err2, html2))

        # Walk the trees and pull out json
        dom1, json1 = self._extract_json(dom1)
        dom2, json2 = self._extract_json(dom2)

        # Convert dom back to string, call super to test doms
        # Yes it's inefficient, but it's tests and saves me from forking it
        super(TagTestManager, self).assertHTMLEqual(
            six.text_type(dom1), six.text_type(dom2),
        )

        # Test jsons
        # Assert we've found the same elements
        self.assertSequenceEqual(json1.keys(), json2.keys())
        for dom_path in json1.keys():
            # Assert the element has the same attributes
            self.assertSequenceEqual(
                json1[dom_path].keys(), json2[dom_path].keys(),
                msg='%s has attributes do not match: %r != %r' % (
                    dom_path, json1[dom_path].keys(), json2[dom_path].keys(),
                )
            )
            for attr_name in json1[dom_path].keys():
                # Django 1.4 doesn't support assertJSONEqual
                # For now, just have a copy of it from 1.5
                try:
                    json1_val = json.loads(json1[dom_path][attr_name])
                except ValueError:
                    self.fail(
                        "%s %s test result is not valid JSON: %r" %
                        (dom_path, attr_name, json1[dom_path][attr_name])
                    )
                else:
                    try:
                        json2_val = json.loads(json2[dom_path][attr_name])
                    except ValueError:
                        self.fail(
                            "%s %s expected is not valid JSON: %r" %
                            (dom_path, attr_name, json2[dom_path][attr_name])
                        )
                    else:
                        self.assertEqual(
                            json1_val, json2_val,
                            msg='%s %s JSON does not match' % (dom_path, attr_name)
                        )

    def debugTagModel(self, model):
        """
        Print tag model tags and their counts, to help debug failed tests
        """
        print("-=-=-=-=-=-")
        if isinstance(model, (tag_models.SingleTagDescriptor, tag_models.TagDescriptor)):
            model = model.tag_model
        print("Tag model: %s" % model)
        for tag in model.objects.all():
            print('%s: %d' % (tag.name, tag.count))
        print("-=-=-=-=-=-")

    def prettyPrint(self, data):
        "Pretty print data, to help debug failed tests"
        import pprint
        pp = pprint.PrettyPrinter(indent=4)
        pp.pprint(data)


# Based on http://stackoverflow.com/questions/16571150/how-to-capture-stdout-output-from-a-python-function-call
class Capturing(list):
    "Capture stdout and stderr to a string"
    def __enter__(self):
        self._stdout = sys.stdout
        self._stderr = sys.stderr
        sys.stdout = sys.stderr = self._stringio = StringIO()
        return self
    def __exit__(self, *args):
        # Ensure output is unicode
        self.extend(
            six.text_type(line)
            for line in self._stringio.getvalue().splitlines()
        )
        sys.stdout = self._stdout
        sys.stderr = self._stderr

def skip_if_mysql(fn):
    if connection.vendor == 'mysql':
        return unittest.skip('MySQL does not support tested feature')(fn)
    return fn



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import DataMigration
from django.db import models

import tagulous


class Migration(DataMigration):
    #
    # Implement basic unittest-esque methods
    #

    def fail(self, msg):
        # Fail loudly so the message can be picked up by the calling test
        raise AssertionError(msg)

    def assertEqual(self, a, b):
        if a != b:
            self.fail("%r != %r" % (a, b))

    def assertIsSubclass(self, cls, supercls):
        if not issubclass(cls, supercls):
            self.fail("%r is not a subclass of %r: %s" % (cls, supercls, cls.__bases__))

    def assertIsInstance(self, obj, cls, msg=None):
        if not isinstance(obj, cls):
            self.fail("%r is not an instance of %r" % (obj, cls))


    #
    # Perform tests
    #

    def forwards(self, orm):
        "Test that Tagulous works in data migrations"
        # Find tagged model
        model = orm['tagulous_tests_migration.MigrationTestModel']

        # Check classes have been assigned correctly
        # If so, everything else will work as it should
        self.assertIsSubclass(model, tagulous.models.TaggedModel)

        self.assertIsInstance(
            model.singletag, tagulous.models.SingleTagDescriptor
        )
        self.assertIsSubclass(
            model.singletag.tag_model, tagulous.models.BaseTagModel
        )

        self.assertIsInstance(
            model.tags, tagulous.models.TagDescriptor
        )
        self.assertIsSubclass(
            model.tags.tag_model, tagulous.models.BaseTagTreeModel
        )


    def backwards(self, orm):
        "Write your backwards methods here."

    models = {
        u'tagulous_tests_migration.tagulous_migrationtestmodel_singletag': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('slug',),)", 'object_name': 'Tagulous_MigrationTestModel_singletag', '_bases': ['tagulous.models.BaseTagModel']},
            'count': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '191'}),
            'protected': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '50'})
        },
        u'tagulous_tests_migration.tagulous_migrationtestmodel_tags': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('slug', 'parent'),)", 'object_name': 'Tagulous_MigrationTestModel_tags', '_bases': ['tagulous.models.BaseTagTreeModel']},
            'count': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'label': ('django.db.models.fields.CharField', [], {'max_length': '191'}),
            'level': ('django.db.models.fields.IntegerField', [], {'default': '1'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '191'}),
            'parent': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'children'", 'null': 'True', 'to': u"orm['tagulous_tests_migration.Tagulous_MigrationTestModel_tags']"}),
            'path': ('django.db.models.fields.TextField', [], {'unique': 'True'}),
            'protected': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '50'})
        },
        u'tagulous_tests_migration.migrationtestmodel': {
            'Meta': {'object_name': 'MigrationTestModel'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '10'}),
            'singletag': ('tagulous.models.fields.SingleTagField', [], {'_set_tag_meta': 'True', 'blank': 'True', 'to': u"orm['tagulous_tests_migration.Tagulous_MigrationTestModel_singletag']", 'null': 'True'}),
            'tags': ('tagulous.models.fields.TagField', [], {'to': u"orm['tagulous_tests_migration.Tagulous_MigrationTestModel_tags']", 'tree': 'True', '_set_tag_meta': 'True'})
        }
    }

    complete_apps = ['tagulous_tests_migration']
    symmetrical = True



"""
Test models
"""
from __future__ import unicode_literals

from django.db import models
from django.utils import six

import tagulous


class MixedModel(models.Model):
    """
    For testing mixed model in a separate app
    """
    name = models.CharField(blank=True, max_length=100)
    singletag = tagulous.models.SingleTagField(
        blank=True, null=True, initial='Mr, Mrs, Ms',
    )
    tags = tagulous.models.TagField(
        blank=True, initial='Adam, Brian, Chris',
    )



"""
Tagulous test app urls

Usage:
    class MyTestCase(TestCase):
        urls = 'tests.tagulous_tests_app.urls'
"""
from __future__ import unicode_literals

import django
from django.conf.urls import include, url

# Django 1.8 removes patterns
if django.VERSION < (1, 8):
    from django.conf.urls import patterns
    mk_urlpatterns = lambda *urls: patterns('', *urls)
else:
    mk_urlpatterns = lambda *urls: list(urls)

import tagulous.views
from tests.tagulous_tests_app import models, views


tagged_model = models.TagFieldOptionsModel

urlpatterns = mk_urlpatterns(
    url(r'^tagulous_tests_app/', include(mk_urlpatterns(
        # CBV with tagged forms
        url(r'views/$', views.null, name='tagulous_tests_app-null'),
        url(r'views/MixedCreate/$',
            views.MixedCreate.as_view(),
            name='tagulous_tests_app-MixedCreate',
        ),
        url(r'views/MixedUpdate/(?P<pk>[0-9]+)/$',
            views.MixedUpdate.as_view(),
            name='tagulous_tests_app-MixedUpdate',
        ),

        # Tagulous autocomplete
        url(r'^autocomplete/unlimited/$',
            tagulous.views.autocomplete,
            {'tag_model': tagged_model.autocomplete_view.tag_model},
            name='tagulous_tests_app-unlimited',
        ),
        url(r'^autocomplete/limited/$',
            tagulous.views.autocomplete,
            {'tag_model': tagged_model.autocomplete_limit.tag_model},
            name='tagulous_tests_app-limited',
        ),
        url(r'^autocomplete/unlimited/login/$',
            tagulous.views.autocomplete_login,
            {'tag_model': tagged_model.autocomplete_view.tag_model},
            name='tagulous_tests_app-login',
        ),
        url(r'^autocomplete/unlimited/queryset/$',
            tagulous.views.autocomplete,
            {'tag_model': tagged_model.autocomplete_view.tag_model.objects.filter(
                name__startswith='tag2',
            )},
            name='tagulous_tests_app-queryset',
        ),
        url(r'^autocomplete/unlimited/force_lowercase_true/$',
            tagulous.views.autocomplete,
            {'tag_model': tagged_model.force_lowercase_true.tag_model},
            name='tagulous_tests_app-force_lowercase_true',
        ),
        url(r'^autocomplete/unlimited/case_sensitive_false/$',
            tagulous.views.autocomplete,
            {'tag_model': tagged_model.case_sensitive_false.tag_model},
            name='tagulous_tests_app-case_sensitive_false',
        ),
        url(r'^autocomplete/unlimited/case_sensitive_true/$',
            tagulous.views.autocomplete,
            {'tag_model': tagged_model.case_sensitive_true.tag_model},
            name='tagulous_tests_app-case_sensitive_true',
        ),
    )))
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.utils import six

import tagulous.models.fields
import tagulous.models.migrations

class Migration(migrations.Migration):

    dependencies = [
        ('tagulous_tests_migration', '0002_tagged'),
    ]

    operations = [
        migrations.AddField(
            model_name='tagulous_migrationtestmodel_tags',
            name='parent',
            field=models.ForeignKey(to='tagulous_tests_migration.Tagulous_MigrationTestModel_tags', related_name='children', blank=True, null=True, on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='tagulous_migrationtestmodel_tags',
            name='label',
            field=models.CharField(default='-', max_length=191, help_text=b'The name of the tag, without ancestors'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='tagulous_migrationtestmodel_tags',
            name='level',
            field=models.IntegerField(default=1, help_text=b'The level of the tag in the tree'),
            preserve_default=True,
        ),
    ] + tagulous.models.migrations.add_unique_field(
        model_name='tagulous_migrationtestmodel_tags',
        name='path',
        field=models.TextField(),
        preserve_default=False,
        set_fn=lambda obj: setattr(obj, 'path', six.text_type(obj.pk)),
    ) + [
        migrations.AlterField(
            model_name='migrationtestmodel',
            name='tags',
            field=tagulous.models.fields.TagField(to='tagulous_tests_migration.Tagulous_MigrationTestModel_tags', help_text=b'Enter a comma-separated tag string', _set_tag_meta=True, tree=True),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='tagulous_migrationtestmodel_tags',
            unique_together=set([('slug', 'parent')]),
        ),
        tagulous.models.migrations.ChangeModelBases(
            name='tagulous_migrationtestmodel_tags',
            bases=(tagulous.models.models.BaseTagTreeModel, models.Model),
        )
    ]



from __future__ import unicode_literals

from django.contrib.auth.decorators import login_required
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models.query import QuerySet
from django.http import HttpResponse
from django.utils import six

# Django 1.4 is last to support Python 2.5, but json isn't available until 2.6
try:
    import json
except ImportError: # pragma: no cover
   from django.utils import simplejson as json


@login_required
def autocomplete_login(*args, **kwargs):
    return autocomplete(*args, **kwargs)

def autocomplete(request, tag_model):
    """
    Arguments:
        request
            The request object from the dispatcher
        tag_model
            Reference to the tag model (eg MyModel.tags.tag_model), or a
            queryset of the tag model (eg MyModel.tags.tag_model.objects.all())

    The following GET parameters can be set:
        q   The query string to filter by (match against start of string)
        p   The current page

    Response is a JSON object with following keys:
        results     List of tags
        more        Boolean if there is more
    }
    """
    # Get model, queryset and tag options
    if isinstance(tag_model, QuerySet):
        queryset = tag_model
        tag_model = queryset.model
    else:
        queryset = tag_model.objects
    options = tag_model.tag_options

    # Get query string
    query = request.GET.get('q', '')
    page = int(request.GET.get('p', 1))

    # Perform search
    if query:
        if options.force_lowercase:
            query = query.lower()

        if options.case_sensitive:
            results = queryset.filter(name__startswith=query)
        else:
            results = queryset.filter(name__istartswith=query)
    else:
        results = queryset.all()

    # Limit results
    if options.autocomplete_limit:
        start = options.autocomplete_limit * (page - 1)
        end = options.autocomplete_limit * page
        more = results.count() > end
        results = results.order_by('name')[start:end]

    # Build response
    response = {
        'results':  [tag.name for tag in results],
        'more':     more,
    }
    return HttpResponse(
        json.dumps(response, cls=DjangoJSONEncoder),
        content_type='application/json',
    )



"""
Tagulous test: manage.py commands

Modules tested:
    tagulous.management.commands.initial_tags
"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from django.core.management import call_command
from django.utils import six

from tests.lib import *


# If True, display output from call_command - use for debugging tests
DISPLAY_CALL_COMMAND = False


###############################################################################
####### ./manage.py initial_tags
###############################################################################

class InitialTagsTest(TagTestManager, TestCase):
    """
    Test initial_tags command
    """
    # Don't list the models in manage_models, otherwise TagTestManager will
    # initialise the tags for us
    def setUpExtra(self):
        self.singletag_model = test_models.SingleTagFieldOptionsModel.initial_list
        self.singletag_model_alt = test_models.SingleTagFieldOptionsModel.initial_string
        self.tag_model = test_models.TagFieldOptionsModel.initial_list
        self.singletag_model_2 = test_models2.MixedModel.singletag
        self.tag_model_2 = test_models2.MixedModel.tags

    def assertModelsEmpty(self):
        self.assertTagModel(self.singletag_model, {})
        self.assertTagModel(self.singletag_model_alt, {})
        self.assertTagModel(self.tag_model, {})
        self.assertTagModel(self.singletag_model_2, {})
        self.assertTagModel(self.tag_model_2, {})

    def assertSingleTagFilled(self, model):
        self.assertTagModel(model, {
            'Mr':   0,
            'Mrs':  0,
            'Ms':   0,
        })
    def assertTagFilled(self, model):
        self.assertTagModel(model, {
            'Adam': 0,
            'Brian': 0,
            'Chris': 0,
        })

    def run_command(self, target=''):
        with Capturing() as output:
            call_command(
                'initial_tags',
                target=target,  # Optional target
                verbosity=1,    # Silent
            )

        if DISPLAY_CALL_COMMAND:
            print(">> initial_tags target=%s" % target)
            print('\n'.join(output))
            print("<<<<<<<<<<")

        return output

    def test_load_all(self):
        "Check no target loads all"
        self.assertModelsEmpty()
        output = self.run_command()

        # Output correct if at least expected fields are loaded (more in apps)
        expected = [
            'tagulous_tests_app.SingleTagFieldOptionsModel.initial_list',
            'tagulous_tests_app.SingleTagFieldOptionsModel.initial_string',
            'tagulous_tests_app.TagFieldOptionsModel.initial_list',
            'tagulous_tests_app2.MixedModel.singletag',
            'tagulous_tests_app2.MixedModel.tags',
        ]
        self.assertGreaterEqual(len(output), len(expected))
        for field in expected:
            self.assertTrue('Loading initial tags for %s' % field in output)

        # All loaded
        self.assertSingleTagFilled(self.singletag_model)
        self.assertSingleTagFilled(self.singletag_model_alt)
        self.assertTagFilled(self.tag_model)
        self.assertSingleTagFilled(self.singletag_model_2)
        self.assertTagFilled(self.tag_model_2)

    def test_load_app(self):
        "Check app target just loads that app"
        self.assertModelsEmpty()
        output = self.run_command('tagulous_tests_app')

        # Output correct if at least expected fields are loaded (more in apps)
        expected = [
            'tagulous_tests_app.SingleTagFieldOptionsModel.initial_list',
            'tagulous_tests_app.SingleTagFieldOptionsModel.initial_string',
            'tagulous_tests_app.TagFieldOptionsModel.initial_list',
        ]
        self.assertGreaterEqual(len(output), len(expected))
        for field in expected:
            self.assertTrue('Loading initial tags for %s' % field in output)

        # Test app 1 loaded
        self.assertSingleTagFilled(self.singletag_model)
        self.assertSingleTagFilled(self.singletag_model_alt)
        self.assertTagFilled(self.tag_model)

        # Test app 2 empty
        self.assertTagModel(self.singletag_model_2, {})
        self.assertTagModel(self.tag_model_2, {})

    def test_load_model(self):
        "Check app and model target just loads that model"
        self.assertModelsEmpty()
        output = self.run_command(
            'tagulous_tests_app.SingleTagFieldOptionsModel'
        )

        # Output correct if at least expected fields are loaded (more in model)
        expected = [
            'tagulous_tests_app.SingleTagFieldOptionsModel.initial_list',
            'tagulous_tests_app.SingleTagFieldOptionsModel.initial_string',
        ]
        self.assertGreaterEqual(len(output), len(expected))
        for field in expected:
            self.assertTrue('Loading initial tags for %s' % field in output)

        # SingleTagFieldOptionsModel loaded
        self.assertSingleTagFilled(self.singletag_model)
        self.assertSingleTagFilled(self.singletag_model_alt)

        # Rest not loaded
        self.assertTagModel(self.tag_model, {})
        self.assertTagModel(self.singletag_model_2, {})
        self.assertTagModel(self.tag_model_2, {})

    def test_load_field(self):
        "Check app, model and field target just loads that field"
        self.assertModelsEmpty()
        output = self.run_command(
            'tagulous_tests_app.SingleTagFieldOptionsModel.initial_list'
        )

        # Output correct if only expected fields are loaded
        expected = [
            'tagulous_tests_app.SingleTagFieldOptionsModel.initial_list',
        ]
        self.assertEqual(len(output), len(expected))
        for field in expected:
            self.assertTrue('Loading initial tags for %s' % field in output)

        # SingleTagFieldOptionsModel loaded
        self.assertSingleTagFilled(self.singletag_model)

        # Rest not loaded
        self.assertTagModel(self.singletag_model_alt, {})
        self.assertTagModel(self.tag_model, {})
        self.assertTagModel(self.singletag_model_2, {})
        self.assertTagModel(self.tag_model_2, {})

    def test_load_field_empty(self):
        "Check loading a field which is empty prints the message"
        self.assertModelsEmpty()
        field = 'tagulous_tests_app.SingleTagFieldOptionsModel.protect_all_true'
        output = self.run_command(field)
        self.assertSequenceEqual(output, [
            'Nothing to load for %s' % field,
        ])
        self.assertModelsEmpty()



"""
Tagulous constants
"""
from __future__ import unicode_literals

from django.utils import six


# Prefix for all models auto-generated by Tagulous
MODEL_PREFIX = 'Tagulous'

# Constants to improve code legibility
COMMA = ','
SPACE = ' '
QUOTE = '"'
DOUBLE_QUOTE = QUOTE + QUOTE
TREE = '/'

# Default model TagField options
OPTION_DEFAULTS = {
    'initial':          '',
    'protect_initial':  True,
    'protect_all':      False,
    'case_sensitive':   False,
    'force_lowercase':  False,
    'max_count':        0,
    'space_delimiter':  True,
    'tree':             False,
    'autocomplete_initial': False,
    'autocomplete_view':    '',
    'autocomplete_limit':   100,
    'autocomplete_settings':    None,
    'get_absolute_url':         None,
    'verbose_name_singular':    None,
    'verbose_name_plural':      None,
}

# List of model TagField options which are relevant to form fields
FORM_OPTIONS = [
    'case_sensitive',
    'force_lowercase',
    'max_count',
    'space_delimiter',
    'tree',
    'autocomplete_limit',
    'autocomplete_settings'
]



from __future__ import unicode_literals

from django.core.management.base import BaseCommand

# Abstract model lookup for django compatibility
try:
    # Django 1.8 and later
    from django.apps import apps
except ImportError:
    # Django 1.7 or earlier
    from django.db.models import get_app, get_models
    get_model = lambda app, model_name: getattr(app, model_name)
else:
    get_app = apps.get_app_config
    get_models = lambda app: app.get_models() if app else apps.get_models()
    get_model = lambda app, model_name: app.get_model(model_name)


from tagulous.models.initial import field_initialise_tags, model_initialise_tags


class Command(BaseCommand):
    """
    Load initial tagulous tags
    Optional argument allows you to specify what to load
    If field_name is missing, initialise all tag fields on the model
    If model_name is missing, initialise all tag fields on models in the app
    If app_name is missing, initialise all tag fields on all models in all apps
    """
    help = 'Load initial tagulous tags'

    def add_arguments(self, parser):
        parser.add_argument(
            '--target',
            type=str,
            default='',
            help='Target to load: [<app_name>[.<model_name>[.<field_name>]]]',
        )

    def handle(self, target='', **options):
        # Split up target argument
        parts = target.split('.')
        app_name, model_name, field_name = parts + ([None] * (3 - len(parts)))

        # Look up app
        if app_name:
            app = get_app(app_name)
        else:
            # get_models(None) will get all models
            app = None

        # Look up specific model, or get all models for the app
        if model_name:
            models = [get_model(app, model_name)]
        else:
            models = get_models(app)

        # If field is specified, can finish here
        if field_name:
            model = models[0]
            field = model._meta.get_field(field_name)
            loaded = field_initialise_tags(
                model, field, report=self.stdout,
            )

            if not loaded:
                self.stdout.write("Nothing to load for %s.%s.%s\n" % (
                    model._meta.app_label,
                    model.__name__,
                    field.name,
                ))
            return

        # Step through all models and load
        for model in models:
            model_initialise_tags(model, report=self.stdout)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.utils import six

import tagulous.models


#
# Basic unittest-esque functions
#

def fail(msg):
    # Fail loudly so the message can be picked up by the calling test
    raise AssertionError(msg)

def assertEqual(a, b):
    if a != b:
        fail("%r != %r" % (a, b))

def assertIsSubclass(cls, supercls):
    if not issubclass(cls, supercls):
        fail("%r is not a subclass of %r: %s" % (cls, supercls, cls.__bases__))

def assertIsInstance(obj, cls, msg=None):
    if not isinstance(obj, cls):
        fail("%r is not an instance of %r" % (obj, cls))


#
# Migration
#

def test_tagulous_in_migrations(apps, schema_editor):
    "Test that Tagulous works in data migrations"
    # Find tagged model
    model = apps.get_model('tagulous_tests_migration', 'MigrationTestModel')

    # Check classes have been assigned correctly
    # If so, everything else will work as it should
    assertIsSubclass(model, tagulous.models.TaggedModel)

    assertIsInstance(
        model.singletag, tagulous.models.SingleTagDescriptor
    )
    assertIsSubclass(
        model.singletag.tag_model, tagulous.models.BaseTagModel
    )

    assertIsInstance(
        model.tags, tagulous.models.TagDescriptor
    )
    assertIsSubclass(
        model.tags.tag_model, tagulous.models.BaseTagTreeModel
    )


class Migration(migrations.Migration):

    dependencies = [
        ('tagulous_tests_migration', '0003_tree'),
    ]

    operations = [
        migrations.RunPython(test_tagulous_in_migrations, lambda *a, **k: None)
    ]



#
# Install the test app and run tests
#
from __future__ import absolute_import
from __future__ import unicode_literals

import warnings

import django
from django.conf import settings
from django.core.management import call_command
from django.utils import six


# Load all tests, if necessary
if django.VERSION < (1, 6):
    from tests.test_options import *
    from tests.test_utils import *
    from tests.test_models_tagmodel import *
    from tests.test_models_singletagfield import *
    from tests.test_models_tagfield import *
    from tests.test_models_order import *
    from tests.test_models_tagged import *
    from tests.test_models_tree import *
    from tests.test_migrations_south import *
    from tests.test_migrations_django import *
    from tests.test_serializers import *
    from tests.test_forms_singletagfield import *
    from tests.test_forms_tagfield import *
    from tests.test_forms_mixed import *
    from tests.test_views import *
    from tests.test_commands import *
    from tests.test_admin import *



"""
Tagulous model field managers

These are accessed via the descriptors, and do the work of storing and loading
the tags.

For tag model manager, look in tagulous.models.models
"""
from __future__ import unicode_literals

import django
from django.core import exceptions
from django.utils import six
from django.utils.encoding import python_2_unicode_compatible

from tagulous.utils import parse_tags, render_tags


###############################################################################
####### Manager for SingleTagField
###############################################################################

class SingleTagManager(object):
    """
    Manage single tags

    Not a real Django manager; it's a per-instance abstraction between the
    normal FK descriptor to hold in-memory changes of the SingleTagField before
    passing them up to the normal FK descriptor on the pre-save signal.
    """
    def __init__(self, descriptor, instance):
        # The SingleTagDescriptor and instance this manages
        self.descriptor = descriptor
        self.instance = instance

        # Other vars we need
        self.tag_model = self.descriptor.tag_model
        self.field = self.descriptor.field
        self.tag_options = self.descriptor.tag_options

        # Keep track of unsaved changes
        self.changed = False

        # The descriptor stores an unsaved tag string
        # Load the actual value into the cache, if it exists
        # If there is a problem with the actual value, get_actual will fall
        # back to our cache, so make sure it exists first
        self.tag_cache = None
        self.tag_cache = self.get_actual()

        # Start off the local tag name with the actual tag name
        self.tag_name = self.tag_cache.name if self.tag_cache else None

        # Pre/post save will need to keep track of an old tag
        self.removed_tag = None

    def flush_cache(self):
        """
        Clear the FK descriptor's cache

        Must be called after all
        """
        # Django 2.0 changes cache management
        if hasattr(self.field, 'get_cached_value'):
            if self.field.is_cached(self.instance):
                self.field.delete_cached_value(self.instance)
        else:
            # Django <2.0
            cache_name = self.field.get_cache_name()
            if hasattr(self.instance, cache_name):
                delattr(self.instance, cache_name)

    def get_actual(self):
        """
        Get the actual value of the instance according to the FK descriptor
        """
        # A ForeignKey would be on the .attname (field_id), but only
        # if it has been set, otherwise the attribute will not exist.
        #
        # In Django <1.9 we could just check for the attribute, but 1.10
        # sets a DeferredAttribute on the field id to load it on demand, so
        # we need to check the dict directly instead
        check_value = self.field.attname in self.instance.__dict__

        # Check the value if we need to
        if check_value:
            try:
                value = self.descriptor.descriptor.__get__(self.instance)
            except (self.tag_model.DoesNotExist, self.instance.DoesNotExist):
                # Django 1.10 returns instance.DoesNotExist, so check for both.
                #
                # If the tag is deleted but nobody tells this instance; because
                # the real cache is kept empty, this will fail. Try our cache,
                # and if it's set clear the pk.
                self.changed = True
                if self.tag_cache:
                    self.tag_cache.pk = None
                return self.tag_cache

            # The descriptor will have populated the cache, but Django 1.8
            # introduces a check of the cache before allowing save() to start.
            # We can't guarantee that the tag object will have a pk until the
            # pre-save handler has fired - which happens after the cache check.
            # We therefore have to keep the cache clean at all times - we don't
            # lose anything since we keep our own cache, and pre-save will fill
            # it out in time anyway.
            self.flush_cache()
            return value
        return None

    def set_actual(self, value):
        """
        Set the actual value of the instance for the FK descriptor
        """
        self.descriptor.descriptor.__set__(self.instance, value)

        # Django 1.8 cache check fix (see comment in get_actual)
        self.flush_cache()

    def get(self):
        """
        Get the current tag - either a Tag object or None
        If the field has been changed since the instance was last saved, the
        Tag object may be a dynamically generated Tag which does not exist in
        the database yet. The count will not be updated until the instance is
        next saved.
        """
        # If changed, find the tag
        if self.changed:
            if not self.tag_name:
                return None

            # Try to look up the tag
            try:
                if self.tag_options.case_sensitive:
                    tag = self.tag_model.objects.get(name=self.tag_name)
                else:
                    tag = self.tag_model.objects.get(name__iexact=self.tag_name)
            except self.tag_model.DoesNotExist:
                # Does not exist yet, create a temporary one (but don't save)
                if not self.tag_cache:
                    self.tag_cache = self.tag_model(name=self.tag_name, protected=False)
                tag = self.tag_cache
            return tag
        else:
            # Return the response that it should have had (a Tag or None)
            return self.get_actual()

    def set(self, value):
        """
        Set the current tag
        """
        if not value:
            tag_name = ''

        elif isinstance(value, six.string_types):
            # Force tag to lowercase
            if self.tag_options.force_lowercase:
                value = value.lower()
            tag_name = value

        # Look up the tag name
        else:
            tag_name = value.name

        # If no change, do nothing
        if self.tag_name == tag_name:
            return

        # Store the tag name and mark changed
        self.changed = True
        self.tag_name = tag_name
        self.tag_cache = None

    def pre_save_handler(self):
        """
        When the model is about to save, update the tag value
        """
        # Get the new tag
        new_tag = self.get()

        # Logic check to replace standard null/blank model field validation
        if not new_tag and self.field.required:
            raise exceptions.ValidationError(self.field.error_messages['null'])

        # Only need to go further if there has been a change
        if not self.changed:
            return

        # Store the old tag so we know to decrement it in post_save
        self.removed_tag = self.get_actual()

        # Create or increment the tag object
        if new_tag:
            # Ensure it is in the database
            if not new_tag.pk:
                new_tag.save()

            # Increment the new tag
            new_tag.increment()

        # Set it
        self.set_actual(new_tag)

        # Clear up
        self.changed = False

    def post_save_handler(self):
        """
        When the model has saved, decrement the old tag
        """
        if self.removed_tag:
            self.removed_tag.decrement()

    def post_delete_handler(self):
        """
        When the model has been deleted, decrement the actual tag
        """
        # Decrement the actual tag
        old_tag = self.get_actual()
        if old_tag:
            old_tag.decrement()
            self.set_actual(None)

            # If there is no new value, mark the old one as a new one,
            # so the database will be updated if the instance is saved again
            if not self.changed:
                self.tag_name = old_tag.name
            self.tag_cache = None
            self.changed = True


###############################################################################
####### Mixin for TagField manager
###############################################################################

@python_2_unicode_compatible
class BaseTagRelatedManager(object):
    """
    Mixin for TagRelatedManagerMixin, and base class for FakeTagRelatedManager.

    Provides methods to managed cached tags
    """
    def init_tagulous(self, descriptor):
        """
        Called directly after the mixin is added to the instantiated manager
        """
        self.tag_model = descriptor.tag_model
        self.tag_options = descriptor.tag_options

        # Maintain an internal set of tags, and track whether they've changed
        # If internal tags are None, haven't been loaded yet
        self.changed = False
        self.tags = None
        self.reload()

    def __str__(self):
        """
        If called on an instance, return the tag string
        """
        return self.get_tag_string()

    def __contains__(self, item):
        item_str = six.text_type(item)
        if self.tag_options.force_lowercase:
            item_str = item_str.lower()

        if self.tag_options.case_sensitive:
            return item_str in [tag.name for tag in self.tags]
        return item_str in [tag.name.lower() for tag in self.tags]

    def __eq__(self, other):
        """
        Compare a tag string or iterable of tags to the tags on this manager
        """
        # If not case sensitive, or lowercase forced, compare on lowercase
        lower = False
        if self.tag_options.force_lowercase or not self.tag_options.case_sensitive:
            lower = True

        # Prep other argument we're comparing against
        if isinstance(other, BaseTagRelatedManager):
            other = other.tags
        if isinstance(other, six.string_types):
            other_str = six.text_type(other)

            # Enforce case non-sensitivity or lowercase
            if lower:
                other_str = other_str.lower()

            # Parse other_str into list of tags
            other_tags = parse_tags(
                other_str, space_delimiter=self.tag_options.space_delimiter,
            )

        else:
            # Assume it's an iterable
            other_tags = other
            if lower:
                other_tags = [six.text_type(tag).lower() for tag in other]

        # Get list of set tags
        self_tags = self.get_tag_list()

        # Compare tag count
        if len(other_tags) != len(self_tags):
            return False

        # Compare tags
        for tag in self_tags:
            # If lowercase or not case sensitive, lower for comparison
            if lower:
                tag = tag.lower()

            # Check tag in other tags
            if tag not in other_tags:
                return False

        # Same number of tags, and all self tags present in other tags
        # It's a match
        return True

    def __ne__(self, other):
        """
        Compare tags, using opposite of __eq__
        """
        return not self.__eq__(other)

    def load_from_tagmanager(self, manager):
        """
        Copy status and cache from the specified manager
        """
        self.changed = manager.changed
        self.tags = manager.tags


    #
    # Functions for getting and setting tag cache
    #

    def get_tag_string(self):
        """
        Get the tag edit string for this instance as a string
        """
        return render_tags(self.tags)

    def get_tag_list(self):
        """
        Get the tag names for this instance as a list of tag names
        """
        return [tag.name for tag in self.tags]

    def set_tag_string(self, tag_string):
        """
        Sets the tags for this instance, given a tag edit string
        """
        # Get all tag names
        tag_names = parse_tags(
            tag_string, space_delimiter=self.tag_options.space_delimiter,
        )

        # Pass on to set_tag_list
        return self.set_tag_list(tag_names)
    set_tag_string.alters_data = True

    def set_tag_list(self, tag_names):
        """
        Sets the tags for this instance, given a list of tag names, or a list
        or queryset of tags
        """
        if self.tag_options.max_count and len(tag_names) > self.tag_options.max_count:
            raise ValueError(
                "Cannot set more than %d tags on this field" %
                self.tag_options.max_count
            )

        # Force tag_names to strings, in case it's a list of tags or a queryset
        tag_names = [six.text_type(tag_name) for tag_name in tag_names]

        # Apply force_lowercase
        if self.tag_options.force_lowercase:
            # Will be lowercase for later comparison
            tag_names = [name.lower() for name in tag_names]

        # Prep tag lookup
        # old_tags      = { cmp_name: tag }
        # cmp_new_names = { cmp_name: cased_name }
        if self.tag_options.case_sensitive:
            old_tags = dict(
                [(tag.name, tag) for tag in self.tags]
            )
            cmp_new_names = dict([(n, n) for n in tag_names])
        else:
            # Not case sensitive - need to compare on lowercase
            old_tags = dict(
                [(tag.name.lower(), tag) for tag in self.tags]
            )
            cmp_new_names = dict(
                [(name.lower(), name) for name in tag_names]
            )

        # See which tags are staying
        new_tags = []
        for cmp_old_name, old_tag in old_tags.items():
            if cmp_old_name in cmp_new_names:
                # Exists - add to new tags
                new_tags.append(old_tag)
                del cmp_new_names[cmp_old_name]
            else:
                # Tag will be removed
                self.changed = True

        # Only left with tag names which aren't present
        for tag_name in cmp_new_names.values():
            # Find or create all new tags
            try:
                if self.tag_options.case_sensitive:
                    tag = self.tag_model.objects.get(name=tag_name)
                else:
                    tag = self.tag_model.objects.get(name__iexact=tag_name)
            except self.tag_model.DoesNotExist:
                # Don't create it until it's saved
                tag = self.tag_model(name=tag_name, protected=False)

            # Add the tag
            new_tags.append(tag)
            self.changed = True

        # Store in internal tag cache
        self.tags = new_tags
    set_tag_list.alters_data = True



class FakeTagRelatedManager(BaseTagRelatedManager):
    """
    Fake manager class to manage cached tags, but provide no database access.

    For use with an unsaved model instance
    """
    _needs_db = '"%r" needs to be saved before TagField can use the database'

    def __init__(self, descriptor, instance, instance_type):
        self.instance = instance
        self.model = instance.__class__
        self.init_tagulous(descriptor)

    def reload(self):
        """
        Instance does not exist in the database, so should wipe the tags
        """
        self.tags = []
        self.changed = False

    def save(self, force=False):
        raise ValueError(self._needs_db % self.instance)

    def set(self, *args):
        raise ValueError(self._needs_db % self.instance)

    def add(self, *args):
        raise ValueError(self._needs_db % self.instance)

    def remove(self, *args):
        raise ValueError(self._needs_db % self.instance)

    def clear(self):
        raise ValueError(self._needs_db % self.instance)


class TagRelatedManagerMixin(BaseTagRelatedManager):
    """
    Mixin for RelatedManager to add tag functions

    Added to the normal m2m RelatedManager, after it has been instantiated.
    This holds in-memory changes of the TagField before committing them to the
    database on the post-save signal.
    """
    def reload(self):
        """
        Get the actual tags
        """
        # Convert to a list to force it to load now, and so we can change it
        self.tags = list(self.all())
        self.changed = False
    reload.alters_data = True

    def save(self, force=False):
        """
        Set the actual tags to the internal tag state

        If force is True, save whether we think it has changed or not
        """
        if not self.changed and not force:
            return

        # Add and remove tags as necessary
        new_tags = self._ensure_tags_in_db(self.tags)
        self.reload()

        # Add new tags
        for new_tag in new_tags:
            if new_tag not in self.tags:
                self.add(new_tag, _enforce_max_count=False)

        # Remove old tags
        for old_tag in self.tags:
            if old_tag not in new_tags:
                self.remove(old_tag)
        self.tags = new_tags
        self.changed = False
    save.alters_data = True

    def _ensure_tags_in_db(self, tags):
        """
        Ensure that self.tags all exist in the database
        """
        db_tags = []
        for tag in tags:
            if tag.pk:
                # Already in DB
                db_tag = tag
            else:
                # Not in DB - get or create
                field_lookup = 'name'
                if not self.tag_options.case_sensitive:
                    field_lookup += '__iexact'
                db_tag, __ = self.tag_model.objects.get_or_create(
                    defaults={'name': tag.name, 'protected': False},
                    **{field_lookup: tag.name}
                )
            db_tags.append(db_tag)
        return db_tags

    #
    # New set, add, remove and clear, to update tag counts
    #
    def set(self, *objs):
        self.clear()
        self.add(*objs)

    def add(self, *objs, **kwargs):
        """
        Add a list of tags or tag strings

        Takes on internal argument, _enforce_max_count - don't use in your code
        """
        enforce_max_count = kwargs.pop('_enforce_max_count', True)
        if kwargs:
            raise TypeError("add() got an unexpected keyword argument")

        # Convert strings to tag objects
        new_tags = []
        for tag in objs:
            if isinstance(tag, six.string_types):
                new_tags.append(self.tag_model(name=tag))
            else:
                new_tags.append(tag)

        # Don't trust the internal tag cache
        self.reload()

        # Reduce tags to ones not already loaded
        new_tags = [
            tag for tag in new_tags
            if tag not in self.tags
        ]

        # Enforce max_count
        if enforce_max_count and self.tag_options.max_count:
            current_count = len(self.tags)
            if current_count + len(new_tags) > self.tag_options.max_count:
                raise ValueError(
                    "Cannot set more than %s tags on this field; it already has %s" % (
                        self.tag_options.max_count, current_count,
                    )
                )

        # Ensure tags exist
        new_tags = self._ensure_tags_in_db(new_tags)

        # Add to db, add to cache, and increment
        super(TagRelatedManagerMixin, self).add(*new_tags)
        for tag in new_tags:
            self.tags.append(tag)
            tag.increment()
    add.alters_data = True

    def remove(self, *objs):
        # Convert strings to tag objects - if object doesn't exist, skip
        rm_tags = []
        for tag in objs:
            if isinstance(tag, six.string_types):
                try:
                    rm_tags.append(self.tag_model.objects.get(name=tag))
                except self.tag_model.DoesNotExist:
                    continue
            else:
                rm_tags.append(tag)

        # Don't trust the internal tag cache
        self.reload()

        # Cut tags back to only ones already set
        rm_tags = [
            tag for tag in self.tags if tag in rm_tags
        ]

        # Remove from cache
        self.tags = [tag for tag in self.tags if tag not in rm_tags]

        # Remove from db and decrement
        super(TagRelatedManagerMixin, self).remove(
            *self._ensure_tags_in_db(rm_tags)
        )
        for tag in rm_tags:
            tag.decrement()
    remove.alters_data = True

    def clear(self):
        # Don't trust the internal tag cache
        self.reload()

        # Clear db, then decrement and empty cache
        super(TagRelatedManagerMixin, self).clear()
        for tag in self.tags:
            tag.decrement()
        self.tags = []
    clear.alters_data = True




from __future__ import unicode_literals

from django.contrib import admin
from django.utils import six

from tests.tagulous_tests_app import models


class SimpleMixedTestAdmin(admin.ModelAdmin):
    """
    For testing tagged models
    For use with models.SimpleMixedTest
    """
    list_display = ('name', 'singletag', 'tags')
    list_filter = ['singletag', 'tags']
    fields = ('name', 'singletag', 'tags')

    # No links for changelist, to simplify tests
    # Django 1.7 supports this being set to None, but Django 1.4 - 1.6 don't
    # Therefore just set to invalid value for now
    list_display_links = ['none']


class SimpleMixedTestInline(admin.TabularInline):
    """
    For testing tagged model inlines on tag model admin
    """
    model = models.SimpleMixedTest
    extra = 3


class SimpleMixedTestSingletagAdmin(admin.ModelAdmin):
    "For testing FK inlines"
    inlines = [SimpleMixedTestInline]


class SimpleMixedTestTagsAdmin(admin.ModelAdmin):
    """
    For testing attribute overrides on TagModels
    For use with models.SimpleMixedTest.tags.tag_model
    """
    list_display = ['name']
    list_filter = ['count']
    exclude = ['name']
    actions = []
    inlines = [SimpleMixedTestInline]




# -*- coding: utf-8 -*-
"""
Django Tagulous - Fabulous Tags
"""
from __future__ import unicode_literals


__version__ = "0.14.1"
__license__ = "BSD"
__author__ = "Richard Terry"
__credits__ = [
    "Pamela McA'Nulty (PamelaM)",
    "James Pic (jpic)",
    "Robert Erb (rerb) at AASHE",
    "Mary (minidietcoke)",
    "Gaël Utard (gutard)",
    "Mark London",
    "Martín R. Guerrero (slackmart)",
    "Peter Baumgartner (ipmb)",
    "Raniere Silva (rgaics)",
    "Dmitry Ivanchenko (ivanchenkodmitry)",
]



"""
Migration support for South and Django migrations
"""
from __future__ import unicode_literals

from tagulous import constants
from tagulous import settings
from tagulous.models.models import BaseTagModel, BaseTagTreeModel
from tagulous.models.tagged import TaggedModel
from tagulous.models.options import TagOptions
from tagulous.models.fields import SingleTagField, TagField
from django.utils import six


###############################################################################
############################################################ Django migrations
###############################################################################

# Check for migration frameworks
try:
    from django.db import migrations as django_migrations
except ImportError:
    # Django migrations not available
    django_migrations = None


def django_support():
    from django.db.migrations import state

    # Monkey-patch Django so it doesn't flatten TagModel abstract base classes.
    # We need them so the BaseTagModel metaclass can set up its options, and
    # so that tag model functionality is available in migrations.
    old_from_model = state.ModelState.from_model.__func__
    def from_model(cls, model, exclude_rels=False):
        base = None
        if issubclass(model, BaseTagTreeModel):
            base = BaseTagTreeModel
        elif issubclass(model, BaseTagModel):
            base = BaseTagModel

        modelstate = old_from_model(cls, model, exclude_rels)

        if base is not None:
            modelstate.bases = (base,) + modelstate.bases
        return modelstate
    state.ModelState.from_model = classmethod(from_model)


# Add Tagulous support to Django migrations if available
if django_migrations is not None:
    django_support()


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#       Helpers for Django migrations
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

if django_migrations is not None:
    class ChangeModelBases(django_migrations.operations.base.Operation):
        """
        Change a Model's bases so it is correct for data migrations
        """
        def __init__(self, name, bases):
            self.name = name
            self.bases = bases

        def state_forwards(self, app_label, state):
            state.models[app_label, self.name.lower()].bases = self.bases

        def database_forwards(self, *args, **kwargs):
            pass
        def database_backwards(self, *args, **kwargs):
            pass


def add_unique_field(model_name, name, field, preserve_default, set_fn):
    """
    Helper for Django migrations which returns a list of Operations to add a
    unique field.

    Warning: only use on a database which supports transactions, in case your
    set_fn method fails and leaves your database in an unusable state.

    Create a migration and provide the string 'x' as the default when prompted.

    Then find the field definition in your migration and replace it with a call
    to this function, removing the default:

        operations = [
            migrations.AddField(
                ...
            ),
            migrations.AddField(
                model_name='mymodel',
                name='my_column',
                field=models.TextField(default='x', unique=True),
                preserve_default=False,
            ),
            migrations.AddField(
                ...
            ),
        ]

    becomes:

        def set_new_column(obj):
            obj.new_column = slugify(obj.name)

        operations = [
            migrations.AddField(
                ...
            ),
        ] + tagulous.models.migration.add_unique_field(
            model_name='mymodel',
            name='my_column',
            field=models.TextField(unique=True),
            preserve_default=False,
            set_fn=set_new_column,
        ) + [
            migrations.AddField(
                ...
            ),
        ]

    Arguments:
        model_name  As django defines
        name        As django defines
        field       As django defines, but without default
        preserve_default    As django defines
        set_fn      Callback to set the field on each instance
    """
    if django_migrations is None: # pragma: no cover
        raise ValueError('Cannot use add_unique_column without Django migrations')

    # Clone the field so it's not unique and can be null
    field_nullable = field.clone()
    field_nullable._unique = False
    field_nullable.null = True

    # RunPython doesn't give the code the app label; pass it as an attribute
    # to save users having to specify the app
    class RunPythonWithAppLabel(django_migrations.RunPython):
        def database_forwards(self, app_label, *args, **kwargs):
            self.code.app_label = app_label
            super(RunPythonWithAppLabel, self).database_forwards(
                app_label, *args, **kwargs
            )

        def database_backwards(self, app_label, *args, **kwargs):
            self.code.app_label = app_label
            super(RunPythonWithAppLabel, self).database_backwards(
                app_label, *args, **kwargs
            )

    # Function to generate the unique value
    def set_unique_values(apps, schema_editor):
        model = apps.get_model(set_unique_values.app_label, model_name)

        # Make sure tag models won't mess with data during this operation
        is_tag_model = issubclass(model, BaseTagModel)
        for obj in model.objects.all():
            set_fn(obj)
            if is_tag_model:
                obj._save_direct()
            else:   # pragma: no cover - no need, _save_direct() calls save()
                obj.save()

    return [
        django_migrations.AddField(
            model_name=model_name,
            name=name,
            field=field_nullable,
            preserve_default=preserve_default,
        ),
        RunPythonWithAppLabel(
            set_unique_values,
            reverse_code=lambda *a, **k: None
        ),
        django_migrations.AlterField(
            model_name=model_name,
            name=name,
            field=field,
            preserve_default=preserve_default,
        ),
    ]


###############################################################################
############################################################ South migrations
###############################################################################

try:
    import south
except ImportError:
    # South not installed
    south = None

def south_support():
    from south import modelsinspector
    from south import orm

    # Monkey-patch South to use BaseTagModel as the base class for tag models
    # in data migrations. It has to use an abstract model without fields,
    # because South adds the fields when it's creating the model, and fields
    # can't be overridden in model subclasses.
    old_get_model_meta = modelsinspector.get_model_meta
    def get_model_meta(model):
        meta_def = old_get_model_meta(model)
        if isinstance(getattr(model, 'tag_options', None), TagOptions):
            if model.tag_options.tree:
                meta_def['_bases'] = ['tagulous.models.BaseTagTreeModel']
            else:
                meta_def['_bases'] = ['tagulous.models.BaseTagModel']
        return meta_def
    modelsinspector.get_model_meta = get_model_meta

    # Monkey-patch South to make sure tagged models subclass TaggedModel.
    # It should happen automatically, and will most of the time - but because
    # South doesn't guarantee order fields are created, it can fail to create
    # ForeignKey and ManyToManyField fields the first time. When that happens
    # it goes back and adds them later - but by then the class_prepared signal
    # will have fired, and we may have missed the tag fields. Therefore we have
    # to re-examine all modules South creates after it has fixed failed fields.
    if settings.ENHANCE_MODELS:
        old_retry_failed_fields = orm._FakeORM.retry_failed_fields
        def retry_failed_fields(self):
            old_retry_failed_fields(self)
            for modelkey, model in self.models.items():
                TaggedModel.cast_class(model)
        orm._FakeORM.retry_failed_fields = retry_failed_fields

    # Build keyword arguments for south
    south_kwargs = {
        # Never want fk
        'to_field':     ['south_supression', {'ignore_if': 'south_supression'}],
        'rel_class':    ['south_supression', {'ignore_if': 'south_supression'}],

        # Never want m2m
        'db_table':     ['south_supression', {'ignore_if': 'south_supression'}],
        'through':      ['south_supression', {'ignore_if': 'south_supression'}],
        'symmetrical':  ['south_supression', {'ignore_if': 'south_supression'}],
    }

    # Add tag options
    for key, value in constants.OPTION_DEFAULTS.items():
        seek = key

        # Store initial as a string
        if key == 'initial':
            seek = 'initial_string'

        # Can't freeze callables - don't need to anyway
        if key == 'get_absolute_url':
            continue

        south_kwargs[key] = ['tag_options.%s' % seek, {'default': value}]

    # Always store _set_tag_meta=True, so migrating tag fields can set tag
    # models' TagMeta during migrations
    south_kwargs['_set_tag_meta'] = [True, {"is_value": True}]

    # Add introspection rule for TagField
    modelsinspector.add_introspection_rules([
        (
            [TagField],     # Class(es) these apply to
            [],             # Positional arguments (not used)
            south_kwargs,   # Keyword arguments
        ),
    ], ["^tagulous\.models\.fields\.TagField"])

    # Create copy of south_kwargs without max_count
    single_south_kwargs = dict(
        (k, v) for k, v in south_kwargs.items() if k != 'max_count'
    )
    modelsinspector.add_introspection_rules([
        (
            [SingleTagField],     # Class(es) these apply to
            [],                   # Positional arguments (not used)
            single_south_kwargs,  # Keyword arguments
        ),
    ], ["^tagulous\.models\.fields\.SingleTagField"])


# Add Tagulous support to South migrations, if available
if south is not None:
    south_support()


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#       Helpers for South
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

def add_unique_column(self, db, model, column, set_fn, field_type, **kwargs):
    """
    Helper for South migrations which add a unique field.

    Warning: only use on a database which supports transactions, in case your
    set_fn method fails and leaves your database in an unusable state.

    This must be the last operation on this table in this migration, otherwise
    when it tries to load data, the orm will not match the database.

    Find the field definition in your migration and replace it with a call to
    this function:

        db.add_column(
            'my_table', 'my_column',
            self.gf(
                'django.db.models.fields.CharField'
            )(default='.', unique=True, max_length=255),
            keep_default=False
        )

    becomes:

        def set_new_column(obj):
            obj.new_column = slugify(obj.name)

        tagulous.models.migrations.add_unique_column(
            self, db, orm['myapp.MyModel'], 'new_column', set_new_column,
            'django.db.models.fields.CharField',
            max_length=255
        )

    Arguments:
        self        Migration object
        db          db object
        model       Model (from orm)
        column      Name of db column to add
        set_fn      Callback to set the field on each instance
        field_type  String reference to Field object
        **kwargs    Arguments for Field object, excluding unique
    """
    if south is None: # pragma: no cover
        raise ValueError('Cannot use add_unique_column without south')

    table = model._meta.db_table

    # Create the column as non-unique
    initial = kwargs.copy()
    initial.update({
        'blank':    True,
        'null':     True,
        'unique':   False,
    })
    db.add_column(
        table, column,
        self.gf(field_type)(**initial),
        keep_default=False
    )

    # Set the fields
    if not db.dry_run:
        # Make sure tag models won't mess with data during this operation
        is_tag_model = issubclass(model, BaseTagModel)
        for obj in model.objects.all():
            set_fn(obj)
            if is_tag_model:
                obj._save_direct()
            else:   # pragma: no cover - no need, _save_direct() calls save()
                obj.save()

    # Change column to unique
    db.alter_column(
        table, column,
        self.gf(field_type)(unique=True, **kwargs)
    )





"""
Tagulous test of serializers, dumpdata and loaddata

Modules tested:
    tagulous.serializers.*
"""
# There were originally separate tests for loaddata and dumpdata, but have been
# combined because in Django 1.4-1.8 serializer order is not deterministic, ie
# order of keys in dicts can change, causing output order to change and
# comparisons to fail on values which are effectively the same. Testing
# dumpdata and loaddata in the same place isn't ideal, but the only other way
# is to implement non-deterministic comparisons for json, pyyaml and xml;
# patches welcome.
#
# This should be fixed in Django 1.9, at which point we can look at splitting
# them again - see https://code.djangoproject.com/ticket/24558

from __future__ import absolute_import
from __future__ import unicode_literals
import os
import tempfile

from django.core import management, serializers
from django.utils import six

from tests.lib import *

try:
    import yaml
except ImportError:
    yaml = None


# Based on django's tests/fixtures/tests.py -> DumpDataAssertMixin
# but without comparisons due to django#24558
class DumpDataAssertMixin(object):

    def _dumpdata(self, args, format='json', filename=None,
                         natural_foreign_keys=False, natural_primary_keys=False,
                         use_base_manager=False, exclude_list=[], primary_keys=''):
        new_io = six.StringIO()
        if filename:
            filename = os.path.join(tempfile.gettempdir(), filename)
        management.call_command('dumpdata', *args, **{'format': format,
                                                      'stdout': new_io,
                                                      'stderr': new_io,
                                                      'output': filename,
                                                      'use_natural_foreign_keys': natural_foreign_keys,
                                                      'use_natural_primary_keys': natural_primary_keys,
                                                      'use_base_manager': use_base_manager,
                                                      'exclude': exclude_list,
                                                      'primary_keys': primary_keys})
        if filename:
            with open(filename, "r") as f:
                command_output = f.read()
            os.remove(filename)
        else:
            command_output = new_io.getvalue().strip()

        # Return early, do comparison externally
        return command_output

        if format == "json":
            self.assertJSONEqual(command_output, output)
        elif format == "xml":
            self.assertXMLEqual(command_output, output)
        else:
            self.assertEqual(command_output, output)


fixture_root = os.path.join(
    os.path.dirname(__file__), 'tagulous_tests_app', 'fixtures'
)


###############################################################################
####### Test serialization of a tagged model
###############################################################################

class SerializationTestMixin(DumpDataAssertMixin):
    """
    Test JSON serializer
    """
    manage_models = [
        test_models.SimpleMixedTest,
    ]
    dump_models = [
        'tagulous_tests_app.Tagulous_SimpleMixedTest_singletag',
        'tagulous_tests_app.Tagulous_SimpleMixedTest_tags',
        'tagulous_tests_app.SimpleMixedTest',
    ]
    fixture_format = 'undefined'

    def setUpExtra(self):
        self.model = test_models.SimpleMixedTest
        self.singletag_model = self.model.singletag.tag_model
        self.tags_model = self.model.tags.tag_model
        self.fixture_name = 'test_fixtures.%s' % self.fixture_format
        self.fixture_path = os.path.join(fixture_root, self.fixture_name)

        # Need to dump to loadable file due to django#24558
        self.tmp_fixture_name = 'tmp_%s_%s' % (testenv, self.fixture_name)
        self.tmp_fixture_path = os.path.join(fixture_root, self.tmp_fixture_name)

        # Delete file if it exists
        if os.path.exists(self.tmp_fixture_path):
            os.remove(self.tmp_fixture_path)

    def tearDownExtra(self):
        if os.path.exists(self.tmp_fixture_path):
            os.remove(self.tmp_fixture_path)

    def _populate(self):
        self.t1 = self.model.objects.create(name='Test 1', singletag='single1', tags='tag1, tag2')
        self.t2 = self.model.objects.create(name='Test 2', singletag='single2', tags='tag3, tag4')

    def _check(self):
        t1 = self.model.objects.get(name='Test 1')
        t2 = self.model.objects.get(name='Test 2')
        self.assertInstanceEqual(t1, name='Test 1', singletag='single1', tags='tag1, tag2')
        self.assertInstanceEqual(t2, name='Test 2', singletag='single2', tags='tag3, tag4')
        self.assertEqual(t1.singletag.name, 'single1')
        self.assertEqual(t2.singletag.name, 'single2')
        self.assertEqual(t1.tags.count(), 2)
        self.assertEqual(t2.tags.count(), 2)
        self.assertTagModel(self.singletag_model, {
            'single1': 1,
            'single2': 1,
        })
        self.assertTagModel(self.tags_model, {
            'tag1': 1,
            'tag2': 1,
            'tag3': 1,
            'tag4': 1,
        })

    def _assert_dumped(self, dumped):
        # Basic checks to make sure it's not completely invalid
        # Cannot test in more detail without django#24558
        self.assertTrue('single1' in dumped)
        self.assertTrue('tag1' in dumped)

    def _empty(self):
        self.t1.delete()
        self.t2.delete()
        self.assertEqual(self.model.objects.count(), 0)
        self.assertTagModel(self.singletag_model, {})
        self.assertTagModel(self.tags_model, {})

    def test_dump_load(self):
        "Check dumpdata and loaddata recreate data"
        # Populate and confirm
        self._populate()
        self._check()

        # Dump
        dumped = self._dumpdata(self.dump_models, format=self.fixture_format)
        self._assert_dumped(dumped)

        # Fixture path
        with open(self.tmp_fixture_path, 'w') as f:
            f.write(dumped)

        # Empty the database
        self._empty()

        # Now load fixture
        management.call_command('loaddata', self.tmp_fixture_name, verbosity=0)

        # Check they were loaded correctly
        self._check()


class JsonSerializationTest(SerializationTestMixin, TagTestManager, TestCase):
    "Test JSON serializer"
    fixture_format = 'json'

@unittest.skipIf(yaml is None, 'pyyaml is not installed')
class YamlSerializationTest(SerializationTestMixin, TagTestManager, TestCase):
    """
    Test YAML serializer
    """
    fixture_format = 'yaml'

class XmlSerializationTest(SerializationTestMixin, TagTestManager, TestCase):
    """
    Test XML serializer
    """
    fixture_format = 'xml'


###############################################################################
####### Check serialization of tagged model with conventional FK and M2M fields
###############################################################################

class MixedTestMixin(SerializationTestMixin):
    manage_models = [
        test_models.MixedNonTagRefTest,
    ]
    dump_models = [
        'tagulous_tests_app.MixedNonTagModel',
        'tagulous_tests_app.MixedNonTagRefTest',
    ]

    def setUpExtra(self):
        self.model = test_models.MixedNonTagRefTest
        self.tag_model = test_models.MixedNonTagModel
        self.fixture_name = 'test_fixtures.%s' % self.fixture_format
        self.fixture_path = os.path.join(fixture_root, self.fixture_name)

        # Need to dump to loadable file due to django#24558
        self.tmp_fixture_name = 'tmp_%s_%s' % (testenv, self.fixture_name)
        self.tmp_fixture_path = os.path.join(fixture_root, self.tmp_fixture_name)

        # Delete file if it exists
        if os.path.exists(self.tmp_fixture_path):
            os.remove(self.tmp_fixture_path)

    def _populate(self):
        tag1 = self.tag_model.objects.create(name='tag1')
        tag2 = self.tag_model.objects.create(name='tag2')
        tag3 = self.tag_model.objects.create(name='tag3')
        tag4 = self.tag_model.objects.create(name='tag4')
        self.t1 = self.model.objects.create(
            name='Test 1', singletag=tag1, tags=[tag1, tag2], fk=tag1
        )
        self.t1.mm.add(tag1, tag2)
        self.t2 = self.model.objects.create(
            name='Test 2', singletag=tag2, tags=[tag3, tag4], fk=tag2
        )
        self.t2.mm.add(tag3, tag4)

    def _check(self):
        t1 = self.model.objects.get(name='Test 1')
        t2 = self.model.objects.get(name='Test 2')
        tag1 = self.tag_model.objects.get(name='tag1')
        tag2 = self.tag_model.objects.get(name='tag2')
        tag3 = self.tag_model.objects.get(name='tag3')
        tag4 = self.tag_model.objects.get(name='tag4')
        self.assertInstanceEqual(
            t1, name='Test 1', singletag='tag1', tags='tag1, tag2',
            fk='tag1', mm=[tag1, tag2]
        )
        self.assertInstanceEqual(
            t2, name='Test 2', singletag='tag2', tags='tag3, tag4',
            fk='tag2', mm=[tag3, tag4]
        )
        self.assertTagModel(self.tag_model, {
            'tag1': 2,
            'tag2': 2,
            'tag3': 1,
            'tag4': 1,
        })

    def _assert_dumped(self, dumped):
        # Basic checks to make sure it's not completely invalid
        # Cannot test in more detail without django#24558
        self.assertTrue('tag1' in dumped)
        self.assertTrue('tag2' in dumped)

    def _empty(self):
        self.t1.delete()
        self.t2.delete()
        self.tag_model.objects.all().delete()
        self.assertEqual(self.model.objects.count(), 0)
        self.assertTagModel(self.tag_model, {})

class MixedJsonSerializationTest(MixedTestMixin, TagTestManager, TestCase):
    "Test JSON serializer with normal fk and m2m fields"
    fixture_format = 'json'

@unittest.skipIf(yaml is None, 'pyyaml is not installed')
class MixedYamlSerializationTest(MixedTestMixin, TagTestManager, TestCase):
    "Test Yaml serializer with normal fk and m2m fields"
    fixture_format = 'yaml'

class MixedXmlSerializationTest(MixedTestMixin, TagTestManager, TestCase):
    "Test XML serializer with normal fk and m2m fields"
    fixture_format = 'xml'


###############################################################################
####### Other serialization tests
###############################################################################

class MixedTestMixin(TagTestManager, TestCase):
    def test_many_to_one(self):
        """
        Check serialization of a tagged model with reverse FKs
        """
        t1 = test_models.MixedRefTest.objects.create(name='test', singletag='test', tags='test')
        rfk1 = test_models.ManyToOneTest.objects.create(name='rfk1', mixed_ref_test=t1)
        # Django 1.7 and earlier don't support refresh_from_db
        t1 = test_models.MixedRefTest.objects.get(pk=t1.pk)
        self.assertEqual(t1.many_to_one.count(), 1)
        # Django 1.5 and earlier don't support .first
        self.assertEqual(t1.many_to_one.all()[0], rfk1)

        serialized = serializers.serialize('xml', test_models.MixedRefTest.objects.all())
        deserialized = list(serializers.deserialize('xml', serialized))
        self.assertEqual(len(deserialized), 1)
        obj = deserialized[0].object
        self.assertInstanceEqual(obj, name='test', singletag='test', tags='test')
        self.assertEqual(obj.many_to_one.count(), 1)
        # Django 1.5 and earlier don't support .first
        self.assertEqual(obj.many_to_one.all()[0].name, 'rfk1')



"""
Tagulous test: Enhanced queryset for tagged models

Will fail if settings.ENHANCE_MODELS is not True

Modules tested:
    tagulous.models.tagged
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from django.core.exceptions import MultipleObjectsReturned
from django.utils import six

from tagulous.models.tagged import _split_kwargs

from tests.lib import *



###############################################################################
####### Test _split_kwargs
###############################################################################

class SplitKwargsText(TagTestManager, TestCase):
    "Test _split_kwargs"
    def setUpExtra(self):
        self.test_model = test_models.MixedTest

    def test_fields(self):
        "Look up known fields"
        safe_fields, singletag_fields, tag_fields = _split_kwargs(
            self.test_model, {
                'name': 'Adam',
                'singletag': 'Mr',
                'tags': 'red, blue',
            }
        )
        self.assertEqual(len(safe_fields), 1)
        self.assertEqual(safe_fields['name'], 'Adam')

        self.assertEqual(len(singletag_fields), 1)
        self.assertEqual(singletag_fields['singletag'], 'Mr')

        self.assertEqual(len(tag_fields), 1)
        self.assertEqual(tag_fields['tags'], 'red, blue')

    def test_with_fields(self):
        "Fields dict returned when with_fields"
        safe_fields, singletag_fields, tag_fields, field_lookup = _split_kwargs(
            self.test_model, {
                'name': 'Adam',
                'singletag': 'Mr',
                'tags': 'red, blue',
            }, with_fields=True,
        )
        self.assertEqual(len(safe_fields), 1)
        self.assertEqual(len(singletag_fields), 1)
        self.assertEqual(len(tag_fields), 1)

        self.assertEqual(len(field_lookup), 3)
        self.assertEqual(
            field_lookup['name'],
            test_models.MixedTest._meta.get_field('name'),
        )
        self.assertEqual(
            field_lookup['singletag'],
            test_models.MixedTest._meta.get_field('singletag'),
        )
        self.assertEqual(
            field_lookup['tags'],
            test_models.MixedTest._meta.get_field('tags'),
        )

    def test_unknown(self):
        "Unknown fields passed as safe"
        safe_fields, singletag_fields, tag_fields = _split_kwargs(
            self.test_model, {
                'unknown':  'field',
            }
        )
        self.assertEqual(len(safe_fields), 1)
        self.assertEqual(safe_fields['unknown'], 'field')
        self.assertEqual(len(singletag_fields), 0)
        self.assertEqual(len(tag_fields), 0)

    def test_lookups(self):
        "Test with lookups enabled"
        safe_fields, singletag_fields, tag_fields = _split_kwargs(
            self.test_model, {
                'name__startswith': 'Ad',
                'singletag__name__contains': 'r',
                'tags__name__contains': 'e',
                'tags__exact': 'red, blue',
            }, lookups=True,
        )

        # Unknown lookups are passed straight
        self.assertEqual(len(safe_fields), 3)
        self.assertEqual(safe_fields['name__startswith'], 'Ad')
        self.assertEqual(safe_fields['singletag__name__contains'], 'r')
        self.assertEqual(safe_fields['tags__name__contains'], 'e')

        self.assertEqual(len(singletag_fields), 0)

        self.assertEqual(len(tag_fields), 1)
        self.assertEqual(tag_fields['tags'], ('red, blue', 'exact'))

    def test_lookups_normal(self):
        "Test with lookups enabled but normal tag"
        safe_fields, singletag_fields, tag_fields = _split_kwargs(
            self.test_model, {
                'tags': 'red',
            }, lookups=True,
        )

        # Unknown lookups are passed straight
        self.assertEqual(len(safe_fields), 0)
        self.assertEqual(len(singletag_fields), 0)

        self.assertEqual(len(tag_fields), 1)
        self.assertEqual(tag_fields['tags'], ('red', None))

    def test_lookups_unknown(self):
        "Unknown fields in lookups passed as safe"
        safe_fields, singletag_fields, tag_fields = _split_kwargs(
            self.test_model, {
                'unknown__exact':  'field',
            }, lookups=True,
        )
        self.assertEqual(len(safe_fields), 1)
        self.assertEqual(safe_fields['unknown__exact'], 'field')
        self.assertEqual(len(singletag_fields), 0)
        self.assertEqual(len(tag_fields), 0)


###############################################################################
####### Test TaggedManager and TaggedQuerySet
###############################################################################

class ModelTaggedQuerysetTest(TagTestManager, TestCase):
    """
    Test enhanced tagged model querysets
    """
    manage_models = [
        test_models.MixedTest,
    ]
    def setUpExtra(self):
        self.test_model = test_models.MixedTest

        self.o1 = self.test_model.objects.create(
            name='Test 1', singletag='Mr', tags='red, green, blue',
        )
        self.o2 = self.test_model.objects.create(
            name='Test 2', singletag='Mrs', tags='red, green, blue',
        )
        self.o3 = self.test_model.objects.create(
            name='Test 3', singletag='Mr', tags='red, green',
        )


    #
    # .get()
    #

    def test_object_get_by_singletag(self):
        "Check that object.get singletag loads the item correctly"
        t1 = self.test_model.objects.get(singletag='Mrs')
        self.assertEqual(t1.pk, self.o2.pk)

    def test_object_get_by_singletag_multiple_raises(self):
        "Check that object.get singletag raises exception when multiple found"
        with self.assertRaises(MultipleObjectsReturned) as cm:
            t1 = self.test_model.objects.get(singletag='Mr')
        self.assertTrue(six.text_type(cm.exception).startswith(
            "get() returned more than one MixedTest -- it returned 2!"
        ))

    def test_object_get_by_tags(self):
        "Check that object.get tags loads the item correctly"
        o4 = self.test_model.objects.create(
            name='Test 4', tags='red, green, yellow',
        )
        t1 = self.test_model.objects.get(tags='red, yellow')
        self.assertEqual(t1.pk, o4.pk)

    def test_object_get_by_tags_multiple_raises(self):
        "Check that object.get tags raises exception when multiple found"
        with self.assertRaises(MultipleObjectsReturned) as cm:
            t1 = self.test_model.objects.get(tags='red, green')
        self.assertTrue(six.text_type(cm.exception).startswith(
            "get() returned more than one MixedTest -- it returned 3!"
        ))

    def test_object_get_by_tags_exact(self):
        "Check that object.get tags loads the item correctly"
        t1 = self.test_model.objects.get(tags__exact='red, green')
        self.assertEqual(t1.pk, self.o3.pk)

    def test_object_get_by_tags_multiple_exact_raises(self):
        "Check that object.get tags raises exception when multiple found"
        o4 = self.test_model.objects.create(
            name='Test 4', tags='yellow, purple',
        )
        o5 = self.test_model.objects.create(
            name='Test 5', tags='yellow, purple',
        )
        with self.assertRaises(MultipleObjectsReturned) as cm:
            t1 = self.test_model.objects.get(tags__exact='yellow, purple')
        self.assertTrue(six.text_type(cm.exception).startswith(
            "get() returned more than one MixedTest -- it returned 2!"
        ))

    def test_object_get_by_singletag_tags(self):
        "Check that object.get singletag and tags loads the item correctly"
        t1 = self.test_model.objects.get(singletag='Mr', tags='red, blue')
        self.assertEqual(t1.pk, self.o1.pk)


    #
    # .get_or_create()
    #

    def test_object_get_or_create_singletagfield(self):
        "Check that objects.get_or_create on SingleTagField creates and gets"
        # Clear out models
        self.test_model.objects.all().delete()
        self.test_model.singletag.tag_model.objects.all().delete()
        self.assertTagModel(self.test_model.singletag, {})

        # Create
        t1, created = self.test_model.objects.get_or_create(
            singletag='Mr',
        )
        self.assertEqual(created, True)
        self.assertTagModel(self.test_model.singletag, {'Mr': 1})
        t1.name = 'Test 4'
        t1.save()
        self.assertEqual(six.text_type(t1.singletag), 'Mr')

        # Get
        t2, created = self.test_model.objects.get_or_create(
            singletag='Mr',
        )
        self.assertEqual(created, False)
        self.assertTagModel(self.test_model.singletag, {'Mr': 1})
        self.assertEqual(t2.name, 'Test 4')
        self.assertEqual(six.text_type(t2.singletag), 'Mr')

        self.assertEqual(t1.pk, t2.pk)

    def test_object_get_or_create_tagfield(self):
        "Check that objects.get_or_create on TagField creates and gets"
        # Clear out models
        self.test_model.objects.all().delete()
        self.test_model.tags.tag_model.objects.all().delete()
        self.assertTagModel(self.test_model.tags, {})

        # Create
        t1, created = self.test_model.objects.get_or_create(
            tags='red, blue',
        )
        self.assertEqual(created, True)
        self.assertTagModel(self.test_model.tags, {'red': 1, 'blue': 1})
        t1.name = 'Test 4'
        t1.save()
        self.assertEqual(six.text_type(t1.tags), 'blue, red')

        # Get
        t2, created = self.test_model.objects.get_or_create(
            tags='red, blue',
        )
        self.assertEqual(created, False)
        self.assertTagModel(self.test_model.tags, {'red': 1, 'blue': 1})
        self.assertEqual(t2.name, 'Test 4')
        self.assertEqual(six.text_type(t1.tags), 'blue, red')

        self.assertEqual(t1.pk, t2.pk)


    #
    # .filter()
    #

    def test_object_singletag_filter(self):
        "Check that object.filter finds the correct items by singletag"
        qs1 = self.test_model.objects.filter(singletag='Mr')
        self.assertEqual(qs1.count(), 2)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o3.pk)

    def test_object_singletag_filter_misses(self):
        "Check that object.filter copes with missing singletag"
        qs1 = self.test_model.objects.filter(singletag='Miss')
        self.assertEqual(qs1.count(), 0)

    def test_object_tags_filter(self):
        "Check that object.filter finds the correct items by tags"
        qs1 = self.test_model.objects.filter(tags='red, blue')
        self.assertEqual(qs1.count(), 2)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o2.pk)

    def test_object_tags_filter_exact(self):
        "Check that object.filter finds the correct items by tags"
        qs1 = self.test_model.objects.filter(tags__exact='red, green, blue')
        self.assertEqual(qs1.count(), 2)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o2.pk)

    def test_object_tags_filter_exact_misses(self):
        "Check that object.filter only returns exact matches"
        qs1 = self.test_model.objects.filter(tags__exact='red, blue')
        self.assertEqual(qs1.count(), 0)

    def test_object_tags_filter_none(self):
        "Check that object.filter returns objects with no tags"
        o4 = self.test_model.objects.create(name='Test 4', tags='')
        qs1 = self.test_model.objects.filter(tags=None)
        self.assertEqual(qs1.count(), 1)
        self.assertEqual(qs1[0].pk, o4.pk)


    #
    # .exclude()
    #

    def test_object_singletag_exclude(self):
        "Check that object.exclude finds the correct items"
        qs1 = self.test_model.objects.exclude(singletag='Mrs')
        self.assertEqual(qs1.count(), 2)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o3.pk)

    def test_object_singletag_exclude_misses(self):
        "Check that object.exclude copes with missing singletag"
        qs1 = self.test_model.objects.exclude(singletag='Miss')
        self.assertEqual(qs1.count(), 3)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o2.pk)
        self.assertEqual(qs1[2].pk, self.o3.pk)

    def test_object_tags_exclude(self):
        "Check that object.exclude excludes when both are matched"
        qs1 = self.test_model.objects.exclude(tags='red, blue')
        self.assertEqual(qs1.count(), 1)
        self.assertEqual(qs1[0].pk, self.o3.pk)

    def test_object_tags_exclude_misses(self):
        "Check that object.exclude copes with missing tag"
        qs1 = self.test_model.objects.exclude(tags='red, yellow')
        self.assertEqual(qs1.count(), 3)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o2.pk)
        self.assertEqual(qs1[2].pk, self.o3.pk)

    def test_object_tags_exclude_exact(self):
        "Check that object.exclude exact excludes by tags"
        qs1 = self.test_model.objects.exclude(tags='red, green, blue')
        self.assertEqual(qs1.count(), 1)
        self.assertEqual(qs1[0].pk, self.o3.pk)

    def test_object_tags_exclude_exact_misses(self):
        "Check that object.exclude only excludes exact matches"
        qs1 = self.test_model.objects.exclude(tags='red, green, yellow')
        self.assertEqual(qs1.count(), 3)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o2.pk)
        self.assertEqual(qs1[2].pk, self.o3.pk)

    def test_object_tags_exclude_none(self):
        "Check that object.exclude returns objects with tags"
        o4 = self.test_model.objects.create(name='Test 4', tags='')
        qs1 = self.test_model.objects.exclude(tags=None)
        self.assertEqual(qs1.count(), 3)
        self.assertEqual(qs1[0].pk, self.o1.pk)
        self.assertEqual(qs1[1].pk, self.o2.pk)
        self.assertEqual(qs1[2].pk, self.o3.pk)


@skip_if_mysql
class ModelTaggedQuerysetOptionsSingleTest(TagTestManager, TestCase):
    """
    Test tag options on tagged model querysets, for SingleTagField

    The only tag options which should be applied to queryset arguments is
    case_sensitive
    """
    manage_models = [
        test_models.SingleTagFieldOptionsModel,
    ]
    def setUpExtra(self):
        self.test_model = test_models.SingleTagFieldOptionsModel
        t1 = self.test_model.objects.create(
            name='Test 1',
            case_sensitive_true='Mr',
            case_sensitive_false='Mr',
        )
        t2 = self.test_model.objects.create(
            name='Test 2',
            case_sensitive_true='mr',
            case_sensitive_false='mr',
        )
        t3 = self.test_model.objects.create(
            name='Test 3',
            case_sensitive_true='Mr',
            case_sensitive_false='Mr',
        )

    def test_setup(self):
        "Confirm setup created objects as expected"
        self.assertTagModel(self.test_model.case_sensitive_true, {
            'Mr':     2,
            'mr':     1,
        })
        self.assertTagModel(self.test_model.case_sensitive_false, {
            'Mr':     3,
        })

    def test_case_sensitive_filter(self):
        "Check case sensitive matches"
        qs1 = self.test_model.objects.filter(case_sensitive_true='Mr')
        self.assertEqual(qs1.count(), 2)
        self.assertEqual(six.text_type(qs1[0].name), 'Test 1')
        self.assertEqual(six.text_type(qs1[1].name), 'Test 3')

    def test_case_sensitive_exclude_matches(self):
        "Check case sensitive excludes"
        qs1 = self.test_model.objects.exclude(case_sensitive_true='Mr')
        self.assertEqual(qs1.count(), 1)
        self.assertEqual(six.text_type(qs1[0].name), 'Test 2')

    def test_case_insensitive_filter(self):
        "Check case insensitive matches"
        qs1 = self.test_model.objects.filter(case_sensitive_false='mr')
        self.assertEqual(qs1.count(), 3)
        self.assertEqual(six.text_type(qs1[0].name), 'Test 1')
        self.assertEqual(six.text_type(qs1[1].name), 'Test 2')
        self.assertEqual(six.text_type(qs1[2].name), 'Test 3')

    def test_case_insensitive_exclude(self):
        "Check case insensitive matches"
        qs1 = self.test_model.objects.exclude(case_sensitive_false='mr')
        self.assertEqual(qs1.count(), 0)


@skip_if_mysql
class ModelTaggedQuerysetOptionsTest(TagTestManager, TestCase):
    """
    Test tag options on tagged model querysets, for TagField

    The only tag options which should be applied to queryset arguments is
    case_sensitive
    """
    manage_models = [
        test_models.TagFieldOptionsModel,
    ]
    def setUpExtra(self):
        self.test_model = test_models.TagFieldOptionsModel
        t1 = self.test_model.objects.create(
            name='Test 1',
            case_sensitive_true='Adam',
            case_sensitive_false='Adam',
        )
        t2 = self.test_model.objects.create(
            name='Test 2',
            case_sensitive_true='adam',
            case_sensitive_false='adam',
        )
        t3 = self.test_model.objects.create(
            name='Test 3',
            case_sensitive_true='Adam',
            case_sensitive_false='Adam',
        )

    def test_setup(self):
        "Confirm setup created objects as expected"
        self.assertTagModel(self.test_model.case_sensitive_true, {
            'Adam':     2,
            'adam':     1,
        })
        self.assertTagModel(self.test_model.case_sensitive_false, {
            'Adam':     3,
        })

    def test_case_sensitive_filter(self):
        "Check case sensitive matches"
        qs1 = self.test_model.objects.filter(case_sensitive_true='Adam')
        self.assertEqual(qs1.count(), 2)
        self.assertEqual(six.text_type(qs1[0].name), 'Test 1')
        self.assertEqual(six.text_type(qs1[1].name), 'Test 3')

    def test_case_sensitive_exclude_matches(self):
        "Check case sensitive excludes"
        qs1 = self.test_model.objects.exclude(case_sensitive_true='Adam')
        self.assertEqual(qs1.count(), 1)
        self.assertEqual(six.text_type(qs1[0].name), 'Test 2')

    def test_case_insensitive_filter(self):
        "Check case insensitive matches"
        qs1 = self.test_model.objects.filter(case_sensitive_false='adam')
        self.assertEqual(qs1.count(), 3)
        self.assertEqual(six.text_type(qs1[0].name), 'Test 1')
        self.assertEqual(six.text_type(qs1[1].name), 'Test 2')
        self.assertEqual(six.text_type(qs1[2].name), 'Test 3')

    def test_case_insensitive_exclude(self):
        "Check case insensitive matches"
        qs1 = self.test_model.objects.exclude(case_sensitive_false='adam')
        self.assertEqual(qs1.count(), 0)


###############################################################################
####### Test TaggedModel
###############################################################################

class ModelTaggedTest(TagTestManager, TestCase):
    """
    Test parts of TaggedModel not caught by other tests
    """
    manage_models = [
        test_models.MixedTest,
    ]
    def setUpExtra(self):
        self.test_model = test_models.MixedTest

    def test_init(self):
        "Check constructor can set singletag and tag fields"
        t1 = self.test_model(
            name='Test 1',
            singletag='Mr',
            tags='red, blue',
        )
        t1.save()
        t2 = self.test_model.objects.get(name='Test 1')
        self.assertEqual(six.text_type(t2.singletag), 'Mr')
        self.assertEqual(six.text_type(t2.tags), 'blue, red')



"""
Tagulous test: Tag options

Modules tested:
    tagulous.options
    tagulous.constants
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from django.utils import six

from tests.lib import *


class TagOptionsTest(TestCase):
    """
    Test TagOptions
    """
    def test_defaults(self):
        opt = tag_models.TagOptions()
        self.assertEqual(opt.items(with_defaults=False), {})
        self.assertEqual(opt.items(), tag_constants.OPTION_DEFAULTS)
        self.assertEqual(opt.form_items(with_defaults=False), {})
        self.assertEqual(opt.form_items(), dict([
            (k, v) for k, v in tag_constants.OPTION_DEFAULTS.items()
            if k in tag_constants.FORM_OPTIONS
        ]))

    def test_override_defaults(self):
        opt = tag_models.TagOptions(
            force_lowercase=True,
            case_sensitive=True,
        )
        local_args = {
            'force_lowercase':  True,
            'case_sensitive':   True,
        }
        self.assertEqual(opt.items(with_defaults=False), local_args)
        expected = {}
        expected.update(tag_constants.OPTION_DEFAULTS)
        expected.update(local_args)
        self.assertEqual(opt.items(), expected)
        self.assertEqual(opt.form_items(with_defaults=False), local_args)
        self.assertEqual(opt.form_items(), dict([
            (k, v) for k, v in expected.items()
            if k in tag_constants.FORM_OPTIONS
        ]))

    def test_initial_none(self):
        opt = tag_models.TagOptions(initial=None)
        self.assertEqual(opt.initial_string, '')
        self.assertEqual(opt.initial, [])

    def test_initial_string(self):
        opt = tag_models.TagOptions(initial='one, two')
        self.assertEqual(opt.initial_string, 'one, two')
        self.assertEqual(opt.initial, ['one', 'two'])

    def test_initial_list(self):
        opt = tag_models.TagOptions(initial=['one', 'two'])
        self.assertEqual(opt.initial_string, 'one, two')
        self.assertEqual(opt.initial, ['one', 'two'])
        self.assertEqual(opt.force_lowercase, False)

    def test_set_invalid(self):
        with self.assertRaises(AttributeError) as cm:
            tag_models.TagOptions(invalid=False)
        self.assertEqual(six.text_type(cm.exception), "invalid")

    def test_get_invalid(self):
        opt = tag_models.TagOptions()
        with self.assertRaises(AttributeError) as cm:
            opt.invalid
        self.assertEqual(six.text_type(cm.exception), "invalid")

    def test_update_dict(self):
        opt = tag_models.TagOptions(initial='Adam, Brian')
        self.assertEqual(opt.initial_string, 'Adam, Brian')

        opt.update({'initial': 'Brian, Chris'})
        self.assertEqual(opt.initial_string, 'Brian, Chris')

    def test_update_object(self):
        opt1 = tag_models.TagOptions(initial='Adam, Brian')
        opt2 = tag_models.TagOptions(initial='Brian, Chris')
        self.assertEqual(opt1.initial_string, 'Adam, Brian')
        self.assertEqual(opt2.initial_string, 'Brian, Chris')

        opt1.update(opt2)
        self.assertEqual(opt1.initial_string, 'Brian, Chris')

    def test_set_missing_dict(self):
        opt = tag_models.TagOptions(initial='Adam, Brian')
        self.assertEqual(opt.initial_string, 'Adam, Brian')
        self.assertEqual(opt.force_lowercase, False)

        opt.set_missing({
            'initial':  'Brian, Chris',
            'force_lowercase': True,
        })
        self.assertEqual(opt.initial_string, 'Adam, Brian')
        self.assertEqual(opt.force_lowercase, True)

    def test_set_missing_object(self):
        opt1 = tag_models.TagOptions(initial='Adam, Brian')
        opt2 = tag_models.TagOptions(initial='Brian, Chris', force_lowercase=True)
        self.assertEqual(opt1.initial_string, 'Adam, Brian')
        self.assertEqual(opt1.force_lowercase, False)
        self.assertEqual(opt2.initial_string, 'Brian, Chris')
        self.assertEqual(opt2.force_lowercase, True)

        opt1.set_missing(opt2)
        self.assertEqual(opt1.initial_string, 'Adam, Brian')
        self.assertEqual(opt1.force_lowercase, True)

    def test_add(self):
        opt1 = tag_models.TagOptions(initial='Adam, Brian')
        opt2 = tag_models.TagOptions(force_lowercase=True)
        self.assertEqual(opt1.initial_string, 'Adam, Brian')
        self.assertEqual(opt1.force_lowercase, False)
        self.assertEqual(opt2.initial_string, '')
        self.assertEqual(opt2.force_lowercase, True)

        opt3 = opt1 + opt2
        self.assertEqual(opt3.initial_string, 'Adam, Brian')
        self.assertEqual(opt3.force_lowercase, True)
        self.assertNotEqual(id(opt1), id(opt2))
        self.assertNotEqual(id(opt2), id(opt3))
        self.assertNotEqual(id(opt3), id(opt1))



"""
Tagulous default settings

Override these by setting new values in your global settings file
"""
from __future__ import unicode_literals

from django.conf import settings
from django.utils import six


#
# Database control settings
#

NAME_MAX_LENGTH = getattr(settings, 'TAGULOUS_NAME_MAX_LENGTH', 255)
SLUG_MAX_LENGTH = getattr(settings, 'TAGULOUS_SLUG_MAX_LENGTH', 50)
LABEL_MAX_LENGTH = getattr(
    settings, 'TAGULOUS_LABEL_MAX_LENGTH', NAME_MAX_LENGTH,
)

# Number of characters to allow for finding a unique slug, ie if set to 5, the
# slug will be truncated by up to 5 characters to allow for a suffix of _9999
SLUG_TRUNCATE_UNIQUE = getattr(settings, 'TAGULOUS_SLUG_TRUNCATE_UNIQUE', 5)


#
# Autocomplete settings
#

AUTOCOMPLETE_JS = getattr(
    settings, 'TAGULOUS_AUTOCOMPLETE_JS', (
        'tagulous/lib/jquery.js',
        'tagulous/lib/select2-3/select2.min.js',
        'tagulous/tagulous.js',
        'tagulous/adaptor/select2-3.js',
    )
)
AUTOCOMPLETE_CSS = getattr(
    settings, 'TAGULOUS_AUTOCOMPLETE_CSS', {
        'all':  ['tagulous/lib/select2-3/select2.css'],
    }
)
AUTOCOMPLETE_SETTINGS = getattr(
    settings, 'TAGULOUS_AUTOCOMPLETE_SETTINGS', None
)

# Admin overrides
ADMIN_AUTOCOMPLETE_JS = getattr(
    settings, 'TAGULOUS_ADMIN_AUTOCOMPLETE_JS', AUTOCOMPLETE_JS
)
ADMIN_AUTOCOMPLETE_CSS = getattr(
    settings, 'TAGULOUS_ADMIN_AUTOCOMPLETE_CSS', AUTOCOMPLETE_CSS
)
ADMIN_AUTOCOMPLETE_SETTINGS = getattr(
    settings, 'TAGULOUS_ADMIN_AUTOCOMPLETE_SETTINGS', AUTOCOMPLETE_SETTINGS
)


#
# Tag weighting defaults, for tag model queryset .weight() method
#

WEIGHT_MIN = getattr(settings, 'TAGULOUS_WEIGHT_MIN', 1)
WEIGHT_MAX = getattr(settings, 'TAGULOUS_WEIGHT_MAX', 6)


#
# Feature flags
#

# Option to automatically enhance Model, Manager and QuerySet classes so they
# know how to work with SingleTagFields and TagFields.
#
# This will automatically change models which use tag fields to subclass
# TaggedModel, their managers to subclass TaggedManager, and their querysets to
# subclass TaggedQuerySet
#
# If this is set to False, certain aspects of tag fields will not work as
# expected; you should consider manually subclassing the relevant classes.
# See settings documentation for more information.
ENHANCE_MODELS = getattr(settings, 'TAGULOUS_ENHANCE_MODELS', True)
